#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/html/htmldbg.cpp,v 1.4 1999/07/11 00:46:39 MJRoberts Exp $";
#endif

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  htmldbg.cpp - debugger helper class implementation
Function
  
Notes
  
Modified
  02/04/98 MJRoberts  - Creation
*/

#include <string.h>

/* include necessary TADS interfaces */
#include <dbg.h>
#include <err.h>
#include <lin.h>
#include <linf.h>
#include <dat.h>
#include <os.h>
#include <bif.h>
#include <obj.h>

#ifndef TADSHTML_H
#include "tadshtml.h"
#endif
#ifndef HTMLDBG_H
#include "htmldbg.h"
#endif
#ifndef HTMLSYS_H
#include "htmlsys.h"
#endif
#ifndef HTMLPRS_H
#include "htmlprs.h"
#endif
#ifndef HTMLFMT_H
#include "htmlfmt.h"
#endif
#ifndef HTMLTAGS_H
#include "htmltags.h"
#endif
#ifndef HTMLDISP_H
#include "htmldisp.h"
#endif
#ifndef HTMLDCFG_H
#include "htmldcfg.h"
#endif

/* the hidden output status variable */
extern "C" int dbghid;

CHtmlDebugHelper::CHtmlDebugHelper()
{
    /* there are no windows in the source window list yet */
    first_srcwin_ = last_srcwin_ = 0;

    /* no current or context line yet */
    cur_is_valid_ = FALSE;
    ctx_is_valid_ = FALSE;
    ctx_stackwin_disp_ = 0;

    /* no stack window yet */
    stack_win_ = 0;

    /* no debug log window yet */
    debuglog_win_ = 0;

    /* no history window yet */
    hist_win_ = 0;

    /* set default tab expansion size of 8 */
    tab_size_ = 8;

    /* no temporary breakpoint yet */
    tmp_bp_valid_ = FALSE;

    /* no breakpoints or source files yet */
    bp_ = 0;
    src_ = 0;

    /* start out with the implicit reference for our caller */
    refcnt_ = 1;

    /* we haven't seen any compiled line source ID's yet */
    last_compiled_line_source_id_ = -1;
    next_line_source_id_ = 0;
}

CHtmlDebugHelper::~CHtmlDebugHelper()
{
    /* close any remaining source windows */
    while (first_srcwin_ != 0)
        first_srcwin_->win_->close_window(TRUE);

    /* delete any remaining special line links */
    delete_line_status(FALSE);

    /* delete internal breakpoint and source lists */
    delete_internal_bps();
    delete_internal_sources();
}

/*
 *   Delete all line links 
 */
void CHtmlDebugHelper::delete_line_status(int redraw)
{
    /* loop through all links, deleting each */
    while (first_line_ != 0)
    {
        CHtmlDbg_line_link *cur;

        /* remember this item */
        cur = first_line_;

        /* remove it from the list */
        first_line_ = first_line_->next_;
        if (first_line_ != 0)
            first_line_->prev_ = 0;

        /* if we're redrawing, update the line to show no status */
        if (redraw)
        {
            /* clear the link's flags */
            cur->stat_ = 0;
            
            /* update it on-screen */
            update_line_on_screen(cur);
        }
        
        /* delete the link */
        delete cur;
    }

    /* there's no more last line either */
    last_line_ = 0;
}

/*
 *   Delete internal breakpoint records 
 */
void CHtmlDebugHelper::delete_internal_bps()
{
    while (bp_ != 0)
    {
        CHtmlDbg_bp *nxt;

        /* remember the next one */
        nxt = bp_->nxt_;

        /* delete this one */
        delete bp_;

        /* move on */
        bp_ = nxt;
    }
}

/*
 *   delete internal source file records 
 */
void CHtmlDebugHelper::delete_internal_sources()
{
    while (src_ != 0)
    {
        CHtmlDbg_src *nxt;

        /* remember the next one */
        nxt = src_->nxt_;

        /* delete this one */
        delete src_;

        /* move on */
        src_ = nxt;
    }

    /* forget all compiled line source ID's */
    last_compiled_line_source_id_ = -1;
    next_line_source_id_ = 0;
}

/*
 *   initialize after loading 
 */
htmldbg_loadbp_stat CHtmlDebugHelper::
   init_after_load(dbgcxdef *ctx,
                   CHtmlDebugSysIfc_win *syswinifc,
                   CHtmlDebugConfig *config, int keep_bp_config)
{
    CHtmlDbg_win_link *link;
    lindef *lin;
    htmldbg_loadbp_stat ret;

    /* presume success */
    ret = HTMLDBG_LBP_OK;

    /* we haven't seen any compiled line source ID's yet */
    last_compiled_line_source_id_ = -1;

    /* 
     *   check to make sure that the correct debug info is available if we
     *   have a game loaded (we might only have a config loaded, in which
     *   case it's too early to determine if the game has the correct
     *   debug information) 
     */
    if (ctx != 0 && !(ctx->dbgcxflg & DBGCXFLIN2))
        errsig(ctx->dbgcxerr, ERR_NEEDLIN2);

    /* 
     *   if they want to keep the current breakpoint configuration, save
     *   the current breakpoint configuration so that we can reload it
     *   into the new TADS engine debug context 
     */
    if (keep_bp_config && config != 0)
        save_bp_config(config);

    /* 
     *   discard all of the old breakpoints - we'll reload them if the
     *   caller wanted to keep them, otherwise they're no longer relevant 
     */
    delete_internal_bps();

    /* 
     *   delete the old line status markers - we'll refresh the display
     *   with a current set when we finish loading 
     */
    delete_line_status(TRUE);

    /* delete the old line source trackers */
    delete_internal_sources();

    /* load the new line sources */
    if (ctx != 0)
    {
        /* read line sources from the compiled game */
        for (lin = ctx->dbgcxlin ; lin != 0 ; lin = lin->linnxt)
        {
            linfdef *linf;
            
            /* create and add a new source tracker for this line */
            linf = (linfdef *)lin;
            add_internal_line_source(linf->linfnam,
                                     linf->linfnam + strlen(linf->linfnam)+1,
                                     lin->linid);
            
            /* if this is the highest ID yet, note it */
            if (lin->linid > last_compiled_line_source_id_)
                last_compiled_line_source_id_ = lin->linid;
        }
    }
    else
    {
        int i;
        
        /* 
         *   We don't have a compiled game available, so read line sources
         *   from the saved configuration. 
         */
        i = config->get_strlist_cnt("srcfiles", "fname");
        while (i > 0)
        {
            textchar_t fname[OSFNMAX];
            textchar_t path[OSFNMAX];

            /* move on to the next one */
            --i;

            /* 
             *   Read this filename, and create the line source.  We saved
             *   each line source in the string array at the index value
             *   equal to its source ID, so the source ID is simply the
             *   current index value ('i').  
             */ 
            if (!config->getval("srcfiles", "fname", i, fname, sizeof(fname))
                && !config->getval("srcfiles", "path", i, path, sizeof(path)))
                add_internal_line_source(fname, path, i);
        }
    }

    /* 
     *   we now know the highest line source ID in the compiled file;
     *   start assigning ID's to external files after this value 
     */
    next_line_source_id_ = last_compiled_line_source_id_ + 1;

    /*
     *   Go through all of the windows and figure out which ones now
     *   contain files that are in our source file list.  For each one
     *   that's in the list, set its source ID.  Even if the game was
     *   previously loaded, the source ID's can change on a recompile, so
     *   we need to fix up all of the ID's each time we reload.  
     */
    for (link = first_srcwin_ ; link != 0 ; link = link = link->next_)
    {
        CHtmlDbg_src *src;

        /* 
         *   if this window isn't tied to a line source, don't bother
         *   trying to tie it to a new line source 
         */
        if (link->source_id_ == -1)
            continue;

        /* see if there's a source file matching this window's file */
        src = find_internal_src(syswinifc, link->fname_.get(),
                                link->path_.get(), TRUE);

        /* 
         *   if it matches, set the source ID to the matching source
         *   file's ID; otherwise, set the source ID to invalid, since
         *   this window contains some random source file that isn't part
         *   of the source code for this game 
         */
        link->source_id_ = (src == 0 ? -1 : src->source_id_);
    }

    /* if they provided a configuration, load breakpoints */
    if (config != 0)
        ret = load_bp_config(ctx, syswinifc, config);

    /* return the result */
    return ret;
}

/*
 *   clear the internal line source list from a configuration object 
 */
void CHtmlDebugHelper::clear_internal_source_config(CHtmlDebugConfig *config)
{
    /* clear out our string lists */
    config->clear_strlist("srcfiles", "fname");
    config->clear_strlist("srcfiles", "path");
}

/*
 *   save the internal line source list 
 */
void CHtmlDebugHelper::save_internal_source_config(CHtmlDebugConfig *config)
{
    CHtmlDbg_src *src;
    
    /* clear out any existing source file list */
    clear_internal_source_config(config);

    /* loop through our internal line sources, saving each one */
    for (src = src_ ; src != 0 ; src = src->nxt_)
    {
        /* save this item; the line source ID is the string list index */
        config->setval("srcfiles", "fname", src->source_id_,
                       src->fname_.get());
        config->setval("srcfiles", "path", src->source_id_,
                       src->path_.get() != 0 ? src->path_.get() : "");
    }
}

/*
 *   add a line source to our internal list
 */
CHtmlDbg_src *CHtmlDebugHelper::
   add_internal_line_source(const textchar_t *fname,
                            const textchar_t *path, int line_source_id)
{
    CHtmlDbg_src *src;

    /* create the new line source object */
    src = new CHtmlDbg_src(fname, path, line_source_id);

    /* 
     *   if this has an assigned ID greater than or equal to the next ID
     *   that we would normally assign a new fake internal line source,
     *   increase our internal assignment counter to ensure we don't
     *   reassign the same ID later
     */
    if (line_source_id >= next_line_source_id_)
        next_line_source_id_ = line_source_id + 1;
    
    /* link it into our list */
    src->nxt_ = src_;
    src_ = src;

    /* return the new line source */
    return src;
}

/*
 *   add an internal line source that doesn't correspond to an actual file
 *   in the compiled game 
 */
CHtmlDbg_src *CHtmlDebugHelper::
   add_fake_internal_line_source(const textchar_t *fname)
{
    int id;
    
    /* assign a new line source ID */
    id = next_line_source_id_++;
    
    /* add the line source */
    return add_internal_line_source(fname, fname, id);
}


/* 
 *   set execution state to GO 
 */
void CHtmlDebugHelper::set_exec_state_go(dbgcxdef *ctx)
{
    /* if there's no context, ignore it */
    if (ctx == 0)
        return;

    /* clear the stepping flags */
    ctx->dbgcxflg &= ~(DBGCXFSS + DBGCXFSO);

    /* remove the old current line pointer while exeucting */
    forget_current_line();
}

/* 
 *   set execution state to STEP OVER - steps without following calls 
 */
void CHtmlDebugHelper::set_exec_state_step_over(dbgcxdef *ctx)
{
    /* if there's no context, ignore it */
    if (ctx == 0)
        return;

    /* set the step and step-over-calls flags */
    ctx->dbgcxflg |= (DBGCXFSS + DBGCXFSO);

    /* keep stepping until we're back at the current depth again */
    ctx->dbgcxsof = ctx->dbgcxdep;

    /* remove the old current line pointer while exeucting */
    forget_current_line();
}

/*
 *   set execution state to STEP OUT - steps until the current function
 *   has returned 
 */
void CHtmlDebugHelper::set_exec_state_step_out(dbgcxdef *ctx)
{
    /* if there's no context, ignore it */
    if (ctx == 0)
        return;

    /* set the step and step-over-calls flags */
    ctx->dbgcxflg |= (DBGCXFSS + DBGCXFSO);

    /* keep stepping until we're back at the *enclosing* depth */
    ctx->dbgcxsof = (ctx->dbgcxdep >= 1 ? ctx->dbgcxdep - 1 : 0);

    /* remove the old current line pointer while exeucting */
    forget_current_line();
}

/* 
 *   set execution state to STEP INTO - steps and follows calls 
 */
void CHtmlDebugHelper::set_exec_state_step_into(dbgcxdef *ctx)
{
    /* if there's no context, ignore it */
    if (ctx == 0)
        return;

    /* set the stepping flag, but clear the step-over-calls flag */
    ctx->dbgcxflg |= DBGCXFSS;
    ctx->dbgcxflg &= ~DBGCXFSO;

    /* remove the old current line pointer while exeucting */
    forget_current_line();
}

/*
 *   Signal a QUIT condition 
 */
void CHtmlDebugHelper::signal_quit(dbgcxdef *ctx)
{
    /* if there's no context, ignore it */
    if (ctx == 0)
        return;

    /* signal the TADS error condition for quitting the game */
    errsig(ctx->dbgcxerr, ERR_RUNQUIT);
}

/*
 *   Signal a RESTART condition 
 */
void CHtmlDebugHelper::signal_restart(dbgcxdef *ctx)
{
    /* if there's no context, ignore it */
    if (ctx == 0)
        return;

    /* 
     *   clear the in-debugger flag - we're leaving the debugger via the
     *   signal, so we don't need to keep this flag set 
     */
    ctx->dbgcxflg &= ~DBGCXFIND;

    /* clear the debugger stack */
    ctx->dbgcxdep = 0;
    ctx->dbgcxfcn = 0;

    /* 
     *   go restart via the built-in restart() function, which takes care
     *   of resetting all the run-time context information 
     */
    bifres((bifcxdef *)ctx->dbgcxrun->runcxbcx, 0);
}

/*
 *   Signal an ABORT condition 
 */
void CHtmlDebugHelper::signal_abort(dbgcxdef *ctx)
{
    /* if there's no context, ignore it */
    if (ctx == 0)
        return;

    /* signal the TADS error condition for quitting the game */
    errsig(ctx->dbgcxerr, ERR_RUNABRT);
}

/*
 *   Move execution to the current selection in the given window.  Returns
 *   zero on success, non-zero on error.  
 */
int CHtmlDebugHelper::set_next_statement(dbgcxdef *ctx,
                                         unsigned int *exec_ofs,
                                         CHtmlSysWin *win)
{
    CHtmlDbg_win_link *win_link;
    unsigned long fpos;
    objnum line_objn;
    uint line_ofs;

    /* if there's no context, we can't do anything */
    if (ctx == 0)
        return 1;

    /* make sure it's valid */
    if (!is_in_same_fn(ctx, exec_ofs, win))
        return 1;

    /* find information on the current line in the window */
    if (srcofs_to_code(ctx, win, &fpos, &win_link, &line_objn, &line_ofs))
        return 2;

    /* move the execution point to the new offset */
    *exec_ofs = line_ofs;

    /* update the current line display */
    forget_current_line();
    set_line_status(win_link->source_id_, fpos, HTMLTDB_STAT_CURLINE);

    /* success */
    return 0;
}

/*
 *   Check if the current selection in the given window is within the same
 *   function or method as the current execution point.  Returns true if
 *   so, false if not.  This is a condition of setting the next statement;
 *   this function is provided so that the UI code can check to make sure
 *   that this condition is met when the user attempts to move the
 *   execution point, and generate an appropriate error message if not.  
 */
int CHtmlDebugHelper::is_in_same_fn(dbgcxdef *ctx,
                                    unsigned int *exec_ofs,
                                    CHtmlSysWin *win)
{
    CHtmlDbg_win_link *win_link;
    unsigned long fpos;
    objnum line_objn;
    uint line_ofs;
    uint start_ofs, end_ofs;
    dbgfdef *fr;

    /* if there's no context, there's no way to compare them */
    if (ctx == 0)
        return FALSE;

    /* get the current frame - if there isn't one, return failure */
    if (ctx->dbgcxfcn == 0)
        return FALSE;
    fr = &ctx->dbgcxfrm[ctx->dbgcxfcn - 1];

    /* find information on the current line in the window */
    if (srcofs_to_code(ctx, win, &fpos, &win_link, &line_objn, &line_ofs))
        return FALSE;

    /* 
     *   if this line is not in the same object as the current target
     *   object, we can disallow it 
     */
    if (line_objn == MCMONINV || line_objn != fr->dbgftarg)
        return FALSE;

    /* 
     *   if there's no property, it's a function, so it's valid as long as
     *   it's in the current object (which we've just determined that it
     *   is) 
     */
    if (fr->dbgfprop == (prpnum)0)
        return TRUE;

    /* find the limits of the property's data range */
    start_ofs = objgetp(ctx->dbgcxmem, line_objn, fr->dbgfprop, 0);
    end_ofs = objgetp_end(ctx->dbgcxmem, line_objn, fr->dbgfprop);

    /* if it's within the range, it's valid, otherwise it's not */
    return (line_ofs >= start_ofs && line_ofs < end_ofs);
}



/*
 *   get the current call trace status 
 */
int CHtmlDebugHelper::is_call_trace_active(dbgcxdef *ctx) const
{
    return ctx == 0 ? FALSE : (ctx->dbgcxflg & DBGCXFTRC) != 0;
}

/*
 *   turn call trace on or off 
 */
void CHtmlDebugHelper::set_call_trace_active(dbgcxdef *ctx, int flag)
{
    /* if there's no context, we can't do anything here */
    if (ctx == 0)
        return;

    /* set or clear the trace flag as appropriate */
    if (flag)
        ctx->dbgcxflg |= DBGCXFTRC;
    else
        ctx->dbgcxflg &= ~DBGCXFTRC;
}

/*
 *   clear the call trace log 
 */
void CHtmlDebugHelper::clear_call_trace_log(dbgcxdef *ctx)
{
    /* if there's no context, ignore it */
    if (ctx == 0)
        return;

    /* forget everything in the buffer */
    ctx->dbgcxhstf = 0;

    /* make sure we reflect this in the display window, if open */
    update_hist_window(ctx);
}

/*
 *   get a pointer to the history log buffer 
 */
const textchar_t *CHtmlDebugHelper::get_call_trace_buf(dbgcxdef *ctx) const
{
    return ctx == 0 ? 0 : ctx->dbgcxhstp;
}

/*
 *   get the size of the data accumulated in the history log 
 */
unsigned long CHtmlDebugHelper::get_call_trace_len(dbgcxdef *ctx) const
{
    return ctx == 0 ? 0 : ctx->dbgcxhstf;
}

/*
 *   Get the object and p-code offset for the current selection in the
 *   given window.  Returns zero on success, non-zero on failure.  On
 *   return, *fpos will be changed to reflect the actual source file
 *   location closest to the input fpos value that contains a line start.
 *   *win_link is filled in with a pointer to the window link structure
 *   for the given window.  
 */
int CHtmlDebugHelper::srcofs_to_code(dbgcxdef *ctx, CHtmlSysWin *win,
                                     unsigned long *linenum,
                                     CHtmlDbg_win_link **win_link,
                                     objnum *line_objn, uint *line_ofs) const
{
    CHtmlDbg_line_link *line_link;

    /* find information the current line in the window */
    *linenum = find_line_info(win, win_link, &line_link);

    /* if we didn't find the window, return failure */
    if (*win_link == 0)
        return 1;

    /* get information on this position */
    return srcofs_to_code(ctx, (*win_link)->source_id_, linenum,
                          line_objn, line_ofs);
}

/*
 *   Find the nearest executable line to a given source location.  If TADS
 *   isn't running, we won't make any adjustment; otherwise, we'll fix up
 *   the line number to the nearest line with executable code, and we'll
 *   fill in *line_objn and *line_ofs with the p-code location of the
 *   code.  
 */
int CHtmlDebugHelper::srcofs_to_code(dbgcxdef *ctx, int source_id,
                                     unsigned long *linenum,
                                     objnum *line_objn, uint *line_ofs) const
{
    lindef *lin;
    uchar posbuf[4];

    /* 
     *   if there's no context, the actual source location is the best we
     *   can do for now - we have no game file to tell us the location of
     *   the next executable line 
     */
    if (ctx == 0)
    {
        *line_objn = MCMONINV;
        *line_ofs = 0;
        return 0;
    }

    /* get the line source */
    lin = CHtmlDbg_win_link::find_lindef(ctx, source_id);
    if (lin == 0)
        return 2;

    /* find the source line at this location */
    oswp4(posbuf, *linenum);
    linfind(lin, (char *)posbuf, line_objn, line_ofs);

    /* 
     *   adjust the file position to the actual value -- this may be
     *   different than the text the user was pointing at, since not every
     *   line contains executable code 
     */
    *linenum = osrp4(posbuf);

    /* make sure we have a valid location */
    if (*line_objn == MCMONINV)
        return 3;

    /* success */
    return 0;
}

/*
 *   Set or clear a breakpoint at the current selection in the given
 *   window 
 */
void CHtmlDebugHelper::toggle_breakpoint(dbgcxdef *ctx,
                                         CHtmlSysWin *win)
{
    CHtmlDbg_win_link *win_link;
    unsigned long linenum;
    objnum line_objn;
    uint line_ofs;

    /* find information on the current line in the window */
    if (srcofs_to_code(ctx, win, &linenum, &win_link, &line_objn, &line_ofs))
        return;

    /* toggle the breakpoint at the given location */
    toggle_breakpoint(ctx, win_link->source_id_, linenum,
                      line_objn, line_ofs, 0, FALSE);
}

/*
 *   Set or clear a breakpoint at a given location.  Returns zero on
 *   success, non-zero on error.  
 */
int CHtmlDebugHelper::toggle_breakpoint(dbgcxdef *ctx,
                                        int source_id, unsigned long linenum,
                                        objnum line_objn, uint line_ofs,
                                        const textchar_t *cond,
                                        int set_only)
{
    int bpnum;
    int did_set;
    char bpname[HTMLDBG_MAXBPNAME];
    CHtmlDbg_src *src;
    CHtmlDbg_bp *bp;

    /* find the internal source file tracker */
    src = find_internal_src(source_id);
    if (src == 0)
        return 1;

    /*
     *   Build the name of the breakpoint for display and configuration
     *   saving purposes.  The name consists of the name of the line
     *   source and the source offset.  
     */
    sprintf(bpname, "#%lu %s", linenum, src->fname_.get());

    /*
     *   If we can only set a new breakpoint, check to see if there's
     *   already a breakpoint at this location.  If there is, disallow it.
     */
    if (set_only && find_internal_bp(source_id, linenum) != 0)
        return 2;

    /* set the breakpoint in the engine, if we're running */
    if (ctx != 0)
    {
        /* set the breakpoint in TADS */
        if (dbgbpat(ctx, line_objn, MCMONINV, line_ofs, &bpnum,
                    bpname, TRUE, (char *)cond, &did_set))
            return 3;
    }

    /* find the internal breakpoint record */
    bp = find_internal_bp(source_id, linenum);

    /* we're setting a new breakpoint if we didn't find an existing one */
    did_set = (bp == 0);

    /* add or remove our internal record */
    if (did_set)
    {
        /* if the engine isn't running, synthesize a breakpoint ID */
        if (ctx == 0)
            bpnum = synthesize_bp_num();
        
        /* create a new breakpoint record */
        bp = new CHtmlDbg_bp(FALSE, source_id, linenum, bpnum, 0);

        /* link it into our list */
        bp->nxt_ = bp_;
        bp_ = bp;
    }
    else
    {
        /* we're clearing the breakpoint - delete our tracker */
        delete_internal_bp(bp);
    }

    /* update the special status of the line */
    change_line_status(source_id, linenum, HTMLTDB_STAT_BP, did_set);

    /* 
     *   if we cleared the breakpoint, make sure we clear any
     *   breakpoint-related flags on the line
     */
    if (!did_set)
        change_line_status(source_id, linenum,
                           (HTMLTDB_STAT_BP_DIS | HTMLTDB_STAT_BP_COND),
                           FALSE);

    /* success */
    return 0;
}

/*
 *   Set a temporary breakpoint at the current selection in the given
 *   window.  If there's already a breakpoint at this location, we won't
 *   add a new breakpoint, and we'll return false.  Otherwise, we'll set a
 *   new breakpoint and return true.
 *   
 *   The purpose of a temporary breakpoint is to allow "run to
 *   cursor"-type operations, where the debugger must secretly set its own
 *   internal breakpoint to facilitate the operation but where the user
 *   doesn't intend to add a new breakpoint to the persistent breakpoint
 *   list.  Because the breakpoint is temporary, we won't bother giving it
 *   a meaningful name for the display list, and we also won't show the
 *   breakpoint in the line status displayed in the source window; this
 *   breakpoint should thus be removed the next time the debugger is
 *   entered.
 *   
 *   Only one temporary breakpoint is allowed at a time.  If there's
 *   already a temporary breakpoint set, we'll remove the old one before
 *   adding the new one.  
 */
int CHtmlDebugHelper::set_temp_bp(dbgcxdef *ctx, CHtmlSysWin *win)
{
    CHtmlDbg_win_link *win_link;
    unsigned long fpos;
    objnum line_objn;
    uint line_ofs;
    int did_set;

    /* if there's no context, ignore it */
    if (ctx == 0)
        return FALSE;

    /* if we already have a temporary breakpoint, clear it */
    clear_temp_bp(ctx);

    /* find information on the current line in the window */
    if (srcofs_to_code(ctx, win, &fpos, &win_link, &line_objn, &line_ofs))
        return FALSE;

    /* if there's already a breakpoint there, there's nothing to do */
    if (dbgisbp(ctx, line_objn, MCMONINV, line_ofs, &tmp_bpnum_))
        return FALSE;

    /* set the breakpoint */
    if (dbgbpat(ctx, line_objn, MCMONINV, line_ofs, &tmp_bpnum_,
                "Temp BP", FALSE, 0, &did_set) || !did_set)
        return FALSE;

    /* 
     *   we've set the breakpoint -- note that we now have a temporary
     *   breakpoint, and return success indication 
     */
    tmp_bp_valid_ = TRUE;
    return TRUE;
}

/*
 *   Clear the temporary breakpoint, if there is one.  If no temporary
 *   breakpoint is currently set, this will have no effect. 
 */
void CHtmlDebugHelper::clear_temp_bp(dbgcxdef *ctx)
{
    /* if we don't have a temporary breakpoint, there's nothing to do */
    if (!tmp_bp_valid_)
        return;

    /* clear the breakpoint */
    dbgbpdel(ctx, tmp_bpnum_);

    /* forget the breakpoint */
    tmp_bp_valid_ = FALSE;
}

/*
 *   Add a new global breakpoint.  Returns a non-zero error code if the
 *   condition is invalid or any other error occurs; returns zero on
 *   success, and fills in *bpnum with the identifier of the new
 *   breakpoint.  
 */
int CHtmlDebugHelper::set_global_breakpoint(dbgcxdef *ctx,
                                            const textchar_t *cond,
                                            int *bpnum)
{
    CHtmlDbg_bp *bp;
    
    /* 
     *   Add the breakpoint in the engine, if we're running.  Use the name
     *   "g" for all global breakpoints; since the breakpoint is based
     *   entirely on the condition string, we don't need to remember
     *   anything about it but that it has a condition string.  
     */
    if (ctx != 0)
    {
        /* set it in the engine */
        if (dbgbpat(ctx, MCMONINV, MCMONINV, 0, bpnum, "g", TRUE,
                    (textchar_t *)cond, 0))
            return 1;
    }
    else
    {
        /* the engine isn't running; synthesize a breakpoint ID */
        *bpnum = synthesize_bp_num();
    }

    /* add an entry to our breakpoint tracker list */
    bp = new CHtmlDbg_bp(TRUE, 0, 0, *bpnum, (textchar_t *)cond);
    bp->nxt_ = bp_;
    bp_ = bp;

    /* success */
    return 0;
}

/*
 *   Enable/disable a breakpoint at the current selection point in the
 *   given window.  Reverses current enabling state.  
 */
void CHtmlDebugHelper::toggle_bp_disable(dbgcxdef *ctx,
                                         CHtmlSysWin *win)
{
    CHtmlDbg_win_link *win_link;
    unsigned long linenum;
    objnum line_objn;
    uint line_ofs;

    /* find information on the current line in the window */
    if (srcofs_to_code(ctx, win, &linenum, &win_link, &line_objn, &line_ofs))
        return;

    /* toggle the status */
    toggle_bp_disable(ctx, win_link->source_id_, linenum,
                      line_objn, line_ofs);
}

/*
 *   enable/disable a breakpoint at the given location 
 */
void CHtmlDebugHelper::toggle_bp_disable(dbgcxdef *ctx,
                                         int source_id, unsigned long linenum,
                                         objnum line_objn, int line_ofs)
{
    CHtmlDbg_bp *bp;
    int bpnum;
    int disabled;
    
    /* 
     *   if the engine is running, ask it if there's a breakpoint there -
     *   if not, there's nothing to do 
     */
    if (ctx != 0)
    {
        if (!dbgisbp(ctx, line_objn, MCMONINV, line_ofs, &bpnum))
            return;

        /* toggle the enabled/disabled state of the breakpoint */
        dbgbpdis(ctx, bpnum, disabled);
        disabled = !dbgisbpena(ctx, bpnum);
    }

    /* find our internal tracker */
    bp = find_internal_bp(source_id, linenum);

    /* if we found it, toggle its state */
    if (bp != 0)
    {
        bp->enabled_ = !bp->enabled_;
        disabled = !bp->enabled_;

        /* update the special status of the line */
        change_line_status(bp->source_id_, bp->linenum_,
                           HTMLTDB_STAT_BP_DIS, disabled);
    }
}


/*
 *   Delete a breakpoint given a breakpoint number 
 */
int CHtmlDebugHelper::delete_breakpoint(dbgcxdef *ctx, int bpnum)
{
    int err;
    CHtmlDbg_bp *bp;

    /* presume no error */
    err = 0;

    /* if TADS is running, tell TADS to delete the breakpoint */
    if (ctx != 0 && (err = dbgbpdel(ctx, bpnum)) != 0)
        return err;

    /* find the breakpoint */
    bp = find_internal_bp(bpnum);

    /* if we don't have an internal tracker for the breakpoint, ignore it */
    if (bp == 0)
        return 1;

    /* update the line status, if it's not a global breakpoint */
    if (!bp->global_)
        change_line_status(bp->source_id_, bp->linenum_,
                           HTMLTDB_STAT_BP, FALSE);

    /* delete our internal tracker */
    delete_internal_bp(bp);

    /* return the result */
    return err;
}

/*
 *   Delete all breakpoints 
 */
void CHtmlDebugHelper::delete_all_bps(dbgcxdef *ctx)
{
    CHtmlDbg_bp *nxt;
    
    /* loop through our breakpoint trackers */
    while (bp_ != 0)
    {
        /* remember the next one, since we're going to delete this one */
        nxt = bp_->nxt_;

        /* delete this one */
        delete_breakpoint(ctx, bp_->bpnum_);

        /* move on */
        bp_ = nxt;
    }
}

/*
 *   Enable or disable a breakpoint, given a breakpoint number 
 */
int CHtmlDebugHelper::enable_breakpoint(dbgcxdef *ctx, int bpnum, int enable)
{
    int err;
    CHtmlDbg_bp *bp;

    /* presume no error */
    err = 0;

    /* find the tracker */
    bp = find_internal_bp(bpnum);

    /* if there's no tracker, it's an error */
    if (bp == 0)
        return 1;

    /* if TADS is running, tell TADS to enable or disable the breakpoint */
    if (ctx != 0 && (err = dbgbpdis(ctx, bpnum, !enable)) != 0)
        return err;

    /* update our tracker */
    bp->enabled_ = enable;

    /* update the line status if appropriate */
    if (!bp->global_)
        change_line_status(bp->source_id_, bp->linenum_,
                           HTMLTDB_STAT_BP_DIS, !enable);

    /* return the result */
    return err;
}

/*
 *   Determine if a given breakpoint is associated with code.  Returns
 *   true if so, false if not. 
 */
int CHtmlDebugHelper::is_bp_at_line(int bpnum)
{
    CHtmlDbg_bp *bp;

    /* find the tracker */
    bp = find_internal_bp(bpnum);

    /* if there's no tracker, it's not at a line */
    if (bp == 0)
        return FALSE;

    /* it's at a line if it's not global */
    return !bp->global_;
}

/*
 *   Go to the code associated with a given breakpoint.  Simply shows the
 *   code; doesn't establish the code's stack context.  Returns the window
 *   associated with the code; returns null if the breakpoint isn't
 *   associated with code.  
 */
CHtmlSysWin *CHtmlDebugHelper::show_bp_code(int bpnum,
                                            CHtmlDebugSysIfc_win *syswinifc)
{
    CHtmlDbg_src *src;
    CHtmlDbg_win_link *link;
    int source_id;
    unsigned long linenum;
    
    /* 
     *   get the line information associated with the breakpoint; if there
     *   isn't any line information available, return failure 
     */
    if (!get_bp_line_info(bpnum, &source_id, &linenum))
        return 0;

    /* get the source file */
    src = find_internal_src(source_id);

    /* if we didn't find the line source, give up */
    if (src == 0)
        return 0;

    /* load or find the source window for this line source */
    link = load_source_file(syswinifc, src);
    if (link == 0)
        return 0;

    /* go to the source line */
    show_source_line(link, linenum);

    /* return the window we found */
    return link->win_;
}

/*
 *   Go to a given line in a source file 
 */
void CHtmlDebugHelper::show_source_line(CHtmlDbg_win_link *link,
                                        unsigned long linenum)
{
    CHtmlDisp *disp;

    /* find the line at this position */
    disp = link->formatter_->find_by_debugsrcpos(linenum);

    /* scroll the source window to show this location */
    if (disp != 0)
    {
        CHtmlRect pos;

        /* get the position of this line, and scroll it into view */
        pos = disp->get_pos();
        link->win_->scroll_to_doc_coords(&pos);

        /* put the selection at teh start of the line */
        link->formatter_->set_sel_range(disp->get_text_ofs(),
                                        disp->get_text_ofs());
    }
}


/*
 *   Enumerate breakpoints in a displayable format 
 */
void CHtmlDebugHelper::
   enum_breakpoints(dbgcxdef *ctx,
                    void (*cbfunc)(void *cbctx, int bpnum,
                                   const textchar_t *desc, int disabled),
                    void *cbctx)
{
    CHtmlDbg_bp *bp;
    CHtmlDbg_bp *nxt;
    
    /* loop over the breakpoints */
    for (bp = bp_ ; bp != 0 ; bp = nxt)
    {
        textchar_t desc[HTMLDBG_MAXBPDESC];

        /* 
         *   TADS can change the enabled status of global breakpoints
         *   without consulting us.  If TADS is running, always update the
         *   status of our breakpoints from the engine.  
         */
        if (ctx != 0)
            bp->enabled_ = dbgisbpena(ctx, bp->bpnum_);
        
        /* 
         *   remember the next one now, before invoking the callback, in
         *   case the callback deletes the current breakpoint (in which
         *   case we will no longer be able to access the structure after
         *   the callback returns) 
         */
        nxt = bp->nxt_;

        /* build the description */
        format_bp_for_display(bp, desc, sizeof(desc));

        /* invoke the callback */
        (*cbfunc)(cbctx, bp->bpnum_, desc, !bp->enabled_);
    }
}

/*
 *   Format a breakpoint description from our internal format to a format
 *   suitable for display to the user.  
 */
void CHtmlDebugHelper::format_bp_for_display(const CHtmlDbg_bp *bp,
                                             textchar_t *buf,
                                             size_t /*buflen*/) const
{
    /* check the breakpoint type */
    if (bp->global_)
    {
        /* it's a global breakpoint - format as "when <condition>" */
        sprintf(buf, "when %s%s", bp->cond_.get(),
                bp->enabled_ ? "" : " [disabled]");
    }
    else
    {
        CHtmlDbg_src *src;
        const textchar_t *cond;
        
        /* it's at a source line - find the source tracker */
        src = find_internal_src(bp->source_id_);

        /* get the condition string */
        cond = bp->cond_.get();
        
        /* format it as "at '<filename>' line <linenum> when <condition>" */
        sprintf(buf, "at '%s' line %ld%s%s%s",
                (src != 0 ? src->fname_.get() : "<unknown source file>"),
                bp->linenum_,
                cond != 0 && cond[0] != '\0' ? " when " : "",
                cond != 0 ? cond : "",
                bp->enabled_ ? "" : " [disabled]");
    }
}


/*
 *   Get a breakpoint's description in a format suitable for display.
 *   Returns zero on success, non-zero on failure.  
 */
int CHtmlDebugHelper::get_bp_desc(int bpnum,
                                  textchar_t *buf, size_t buflen)
{
    CHtmlDbg_bp *bp;

    /* find the breakpoint */
    bp = find_internal_bp(bpnum);

    /* if we didn't find it, return failure */
    if (bp == 0)
        return 1;

    /* format the text into the caller's buffer */
    format_bp_for_display(bp, buf, buflen);

    /* success */
    return 0;
}

/*
 *   Determine if a breakpoint is enabled or disabled.  Returns true if
 *   the breakpoint is enabled, false if it's disabled.  
 */
int CHtmlDebugHelper::is_bp_enabled(int bpnum)
{
    CHtmlDbg_bp *bp;

    /* find the breakpoint and return its status */
    bp = find_internal_bp(bpnum);
    return (bp == 0 ? FALSE : bp->enabled_);
}

/*
 *   Get a breakpoint's condition.  Returns zero on success, non-zero on
 *   failure.  If this is a valid breakpoint without a condition, we'll
 *   return success and an empty condbuf string.  
 */
int CHtmlDebugHelper::get_bp_cond(int bpnum,
                                  textchar_t *condbuf, size_t condbuflen)
{
    CHtmlDbg_bp *bp;
    size_t len;

    /* find the breakpoint */
    bp = find_internal_bp(bpnum);
    if (bp == 0)
        return 1;

    /* if the condition doesn't fit, return failure */
    len = (bp->cond_.get() != 0 ? get_strlen(bp->cond_.get()) : 0);

    /* if it doesn't fit, return an error */
    if (len + 1 > condbuflen)
        return 2;

    /* copy the condition */
    if (len != 0)
        do_strcpy(condbuf, bp->cond_.get());
    else
        condbuf[0] = '\0';

    /* success */
    return 0;
}

/*
 *   Change a breakpoint's condition 
 */
int CHtmlDebugHelper::change_bp_cond(dbgcxdef *ctx, int bpnum,
                                     const textchar_t *cond)
{
    CHtmlDbg_bp *bp;
    
    /* if TADS is running, update its condition text */
    if (ctx != 0
        && dbgbpsetcond(ctx, bpnum, (textchar_t *)cond))
        return 1;

    /* update our internal tracker's condition string */
    bp = find_internal_bp(bpnum);
    if (bp != 0)
        bp->cond_.set(cond);

    /* success */
    return 0;
}

/*
 *   Clear the breakpoint configuration 
 */
void CHtmlDebugHelper::clear_bp_config(CHtmlDebugConfig *config) const
{
    /* delete any existing configuration information for the breakpoints */
    config->clear_strlist("breakpoints", "global");
    config->clear_strlist("breakpoints", "file");
    config->clear_strlist("breakpoints", "line");
    config->clear_strlist("breakpoints", "status");
    config->clear_strlist("breakpoints", "cond");
}

/*
 *   Save breakpoint configuration 
 */
void CHtmlDebugHelper::save_bp_config(CHtmlDebugConfig *config) const
{
    CHtmlDbg_bp *bp;
    int i;

    /* clear any existing configuration */
    clear_bp_config(config);

    /* loop over breakpoints */
    for (i = 0, bp = bp_ ; bp != 0 ; bp = bp->nxt_, ++i)
    {
        char buf[20];
        CHtmlDbg_src *src;

        /* if this is a source breakpoint, get the line source */
        src = (bp->global_ ? 0 : find_internal_src(bp->source_id_));

        /* save this breakpoint's global/local status */
        config->setval("breakpoints", "global", i,
                       bp->global_ ? "g" : "l");
        
        /* save the filename */
        config->setval("breakpoints", "file", i,
                       src != 0 ? src->fname_.get() : "");

        /* save the line number */
        sprintf(buf, "%lu", bp->linenum_);
        config->setval("breakpoints", "line", i, buf);

        /* save it's enabled/disabled status */
        config->setval("breakpoints", "status", i,
                       bp->enabled_ ? "e" : "d");
        
        /* if there's a condition, save it */
        config->setval("breakpoints", "cond", i,
                       bp->cond_.get() != 0 ? bp->cond_.get() : "");
    }
}

/*
 *   Load breakpoint configuration 
 */
htmldbg_loadbp_stat CHtmlDebugHelper::
   load_bp_config(dbgcxdef *dbgctx,
                  CHtmlDebugSysIfc_win *syswinifc,
                  CHtmlDebugConfig *config)
{
    int cnt;
    int i;
    int not_set_cnt;
    int moved_cnt;

    /* we haven't failed to set any yet */
    not_set_cnt = 0;

    /* none of the breakpoints have moved yet */
    moved_cnt = 0;
    
    /* run through the breakpoints in the configuration information */
    cnt = config->get_strlist_cnt("breakpoints", "global");
    for (i = 0 ; i < cnt ; ++i)
    {
        textchar_t glob[5];
        textchar_t fname[HTMLDBG_MAXBPNAME];
        textchar_t linebuf[20];
        textchar_t stat[5];
        textchar_t cond[HTMLDBG_MAXCOND];
        
        /* load this breakpoint's information */
        if (config->getval("breakpoints", "global", i, glob, sizeof(glob))
            || config->getval("breakpoints", "file", i, fname, sizeof(fname))
            || config->getval("breakpoints", "line",
                              i, linebuf, sizeof(linebuf))
            || config->getval("breakpoints", "status", i, stat, sizeof(stat))
            || config->getval("breakpoints", "cond", i, cond, sizeof(cond)))
        {
            ++not_set_cnt;
            continue;
        }

        /* create the appropriate type of breakpoint */
        if (glob[0] == 'l')
        {
            unsigned long linenum;
            unsigned long orig_linenum;
            objnum line_objn;
            uint line_ofs;
            CHtmlDbg_src *src;
            
            /* it's a breakpoint at a source line - get the line number */
            orig_linenum = linenum = (unsigned long)atol(linebuf);

            /* find the source file */
            src = find_internal_src(syswinifc, fname, 0, FALSE);
            if (src == 0)
            {
                /* 
                 *   if we have a debugger context, note that this one
                 *   can't actually be set; if not, the game isn't yet
                 *   running, so we're not actually trying to set any
                 *   breakpoints, hence we can't fail to set any
                 *   breakpoints 
                 */
                if (dbgctx != 0)
                    ++not_set_cnt;

                /* create an internal line source for the file */
                src = add_fake_internal_line_source(fname);
            }

            /* if TADS is running, get the exact breakpoint location */
            if (dbgctx != 0
                && srcofs_to_code(dbgctx, src->source_id_, &linenum,
                                  &line_objn, &line_ofs))
            {
                ++not_set_cnt;
                continue;
            }

            /* 
             *   if the line location changed (which can happen if the
             *   game was recompiled since the configuration was last
             *   saved, since some code could have moved around in the
             *   source file), note it 
             */
            if (linenum != orig_linenum)
                ++moved_cnt;

            /* set the breakpoint */
            if (toggle_breakpoint(dbgctx, src->source_id_, linenum,
                                  line_objn, line_ofs, cond, TRUE))
            {
                ++not_set_cnt;
                continue;
            }

            /* if it's disabled, disable it */
            if (stat[0] == 'd')
                toggle_bp_disable(dbgctx, src->source_id_, linenum,
                                  line_objn, line_ofs);
        }
        else
        {
            int bpnum;
            
            /* it's a global breakpoint - set it */
            if (set_global_breakpoint(dbgctx, cond, &bpnum) == 0)
            {
                /* disable the breakpoint if appropriate */
                if (stat[0] == 'd')
                    enable_breakpoint(dbgctx, bpnum, FALSE);
            }
        }
    }

    /* return the result */
    if (not_set_cnt != 0)
        return HTMLDBG_LBP_NOT_SET;
    else if (moved_cnt != 0)
        return HTMLDBG_LBP_MOVED;
    else
        return HTMLDBG_LBP_OK;
}

/*
 *   callback structure for eval_expr_cb 
 */
struct eval_expr_cb_ctx_t
{
    /* debugger context */
    struct dbgcxdef *dbgctx;
    
    /* pointer to start of buffer and original length */
    textchar_t *buf;
    size_t buflen;

    /* pointer to the current position in the output buffer */
    textchar_t *outp;

    /* number of characters remaining in output buffer */
    size_t rem;

    /* logged error number */
    int err;

    /* flag indicating whether we want to keep logged error messages */
    int log_errors;
};

/*
 *   Get the text of an error message.  This routine doesn't require an
 *   error stack, so it's suitable for retrieving the text of error
 *   messages returned from the debugger API. 
 */
void CHtmlDebugHelper::get_error_text(struct dbgcxdef *ctx, int err,
                                      textchar_t *buf, size_t buflen)
{
    /* if there's no context, we can't do anything */
    if (ctx == 0)
    {
        buf[0] = '\0';
        return;
    }
    
    /* get the text of the message */
    errmsg(ctx->dbgcxerr, buf, buflen, err);
}

/*
 *   Get the text of an error message, formatting the message with
 *   arguments from the error stack.  This should only be used when a
 *   valid error stack is present, such as when entering the debugger due
 *   to a run-time error.  
 */
void CHtmlDebugHelper::format_error(struct dbgcxdef *ctx, int err,
                                    textchar_t *buf, size_t buflen)
{
    errcxdef *errctx = ctx->dbgcxerr;
    char msgtxt[256];

    /* get the text of the message */
    errmsg(errctx, msgtxt, sizeof(msgtxt), err);

    /* format it with arguments from the error stack */
    errfmt(buf, buflen, msgtxt,
           errctx->errcxptr->erraac, errctx->errcxptr->erraav);
}

/*
 *   Evaluate an expression 
 */
int CHtmlDebugHelper::eval_expr(struct dbgcxdef *ctx,
                                textchar_t *buf, size_t buflen,
                                const textchar_t *expr, int *is_lval,
                                int *is_openable,
                                void (*aggcb)(void *, const char *, int,
                                              const char *),
                                void *aggctx, int speculative)
{
    int err;
    int level;
    eval_expr_cb_ctx_t cbctx;
    void (*old_errcxlog)(void *, char *, int, int, erradef *);
    void  *old_errcxlgc;
    dattyp dat;

    /* we can't proceed if there's no context */
    if (ctx == 0)
    {
        buf[0] = '\0';
        return 1;
    }

    /* 
     *   Set up the callback context to point to the output buffer.
     *   Reserve one character of the buffer for null termination. 
     */
    cbctx.dbgctx = ctx;
    cbctx.outp = cbctx.buf = buf;
    cbctx.rem = cbctx.buflen = buflen - 1;
    cbctx.err = 0;
    cbctx.log_errors = TRUE;

    /* if we're not in the debugger, we can't evaluate the expression */
    if (!(ctx->dbgcxflg & DBGCXFIND))
    {
        /* put the message in the buffer */
        err = ERR_DBGINACT;
        errmsg(ctx->dbgcxerr, buf, buflen, err);

        /* return the error */
        return err;
    }

    /* 
     *   get the frame level - note that our stack level counter has level
     *   0 as the innermost (current) level, which is the opposite of the
     *   way that dbgeval() counts levels, so we need to invert our
     *   counter to get the one to use in the dbgeval() call 
     */
    if (ctx_stackwin_disp_ != 0)
        level = ctx->dbgcxfcn - ctx_stackwin_disp_->get_debugsrcpos() - 1;
    else
        level = ctx->dbgcxfcn - 1;

    /* make sure we have a valid level, in case we have no stack frames */
    if (level < 0)
    {
        err = ERR_INACTFR;
        errmsg(ctx->dbgcxerr, buf, buflen, err);
        return err;
    }

    /*
     *   While we're evaluating, set up to receive error logging messages
     *   -- we'll write them to our output buffer and note the errors.
     *   Certain expression evaluation errors result in logged (rather
     *   than signalled) errors, so we need to trap any logged errors.  
     */
    old_errcxlog = ctx->dbgcxerr->errcxlog;
    old_errcxlgc = ctx->dbgcxerr->errcxlgc;
    ctx->dbgcxerr->errcxlog = eval_expr_errlog;
    ctx->dbgcxerr->errcxlgc = &cbctx;

    /* evaluate the expression */
    err = dbgevalext(ctx, (char *)expr, eval_expr_cb, &cbctx, level, FALSE,
                     &dat, aggcb, aggctx, speculative);

    /*
     *   determine whether the type is openable, and inform the caller if
     *   desired 
     */
    if (is_openable != 0)
    {
        switch(dat)
        {
        case DAT_LIST:
        case DAT_OBJECT:
            /* these types are openable */
            *is_openable = TRUE;
            break;

        default:
            /* other types are not openable */
            *is_openable = FALSE;
            break;
        }
    }

    /* null-terminate the value buffer */
    *(cbctx.outp) = '\0';

    /* check for error */
    if (err != 0)
    {
        char msg[256];
        
        /* an error occurred - put its text into the buffer */
        errmsg(ctx->dbgcxerr, msg, sizeof(msg), err);
        errfmt(buf, buflen, msg, 0, 0);
    }
    else if (cbctx.err != 0)
    {
        /* a logged error occurred; it's already in the buffer */
        err = cbctx.err;
    }
    else
    {
        /*
         *   If the caller wants to know, check the expression to
         *   determine if it's an lvalue.  To do this, try parsing "(expr)
         *   := 0"; if the expression is an lvalue, the compiler will
         *   accept this, otherwise a parsing error will occur.  
         */
        if (is_lval != 0)
        {
            char asibuf[1024];
            objnum objn;

            /* 
             *   this parse will be speculative, so don't keep any error
             *   messages that get logged 
             */
            cbctx.log_errors = FALSE;

            /* build the assignment and evaluate it */
            sprintf(asibuf, "(%s):=0", expr);
            objn = MCMONINV;
            if (dbgcompile(ctx, asibuf, &ctx->dbgcxfrm[level], &objn,
                           speculative) != 0
                || cbctx.err != 0)
            {
                /* it's not an lvalue, so it can't be assigned into */
                *is_lval = FALSE;
            }
            else
            {
                /* it can be used as an lvalue */
                *is_lval = TRUE;
            }

            /*   
             *   free the compiled code - we didn't need it except for
             *   parsing it 
             */
            if (objn != MCMONINV)
                mcmfre(ctx->dbgcxmem, objn);
        }
    }

    /* if an error occurred, it's definitely not an lvalue */
    if (err != 0 && is_lval != 0)
        *is_lval = FALSE;

    /* restore error context */
    ctx->dbgcxerr->errcxlog = old_errcxlog ;
    ctx->dbgcxerr->errcxlgc = old_errcxlgc;

    /* return the result */
    return err;
}

/*
 *   Display callback for eval_expr - simply adds text to the output
 *   buffer 
 */
void CHtmlDebugHelper::eval_expr_cb(void *ctx0, char *txt, int len)
{
    eval_expr_cb_ctx_t *ctx = (eval_expr_cb_ctx_t *)ctx0;

    /* if an error has been logged, don't record any more output */
    if (ctx->err != 0)
        return;

    /* if this is a single newline, omit it */
    if (len == 1 && *txt == '\n')
        return;

    /* limit the copy to the amount of room we have */
    if ((size_t)len > ctx->rem)
        len = (int)ctx->rem;

    /* copy the characters, turning newlines into spaces */
    while (len > 0)
    {
        *(ctx->outp++) = (*txt == '\n' ? ' ' : *txt);
        ++txt;
        --(ctx->rem);
        --len;
    }
}

/*
 *   Error logging callback for expression evaluator.  We'll put the error
 *   into our buffer and note the error. 
 */
void CHtmlDebugHelper::eval_expr_errlog(void *ctx0, char * /*fac*/, int err,
                                        int argc, erradef *argv)
{
    char msg[256];
    eval_expr_cb_ctx_t *ctx = (eval_expr_cb_ctx_t *)ctx0;

    /* keep the message only if we're currently logging errors */
    if (ctx->log_errors)
    {
        /* format the message into the buffer at the start of the buffer */
        errmsg(ctx->dbgctx->dbgcxerr, msg, sizeof(msg), err);
        errfmt(ctx->buf, ctx->buflen, msg, argc, argv);
        
        /* position the write pointer after the message */
        ctx->outp = ctx->buf + get_strlen(ctx->buf);
        ctx->rem = ctx->buflen - get_strlen(ctx->buf);
    }

    /* record the error */
    ctx->err = err;
}

/*
 *   Enumerate local variables in the current stack context.  
 */
void CHtmlDebugHelper::
   enum_locals(dbgcxdef *ctx,
               void (*cbfunc)(void *, const textchar_t *, size_t),
               void *cbctx)
{
    int level;
    
    /* if we're not in the debugger, there's no local variable context */
    if (ctx == 0 || !(ctx->dbgcxflg & DBGCXFIND))
        return;

    /* get the frame level */
    if (ctx_stackwin_disp_ != 0)
        level = ctx->dbgcxfcn - ctx_stackwin_disp_->get_debugsrcpos() - 1;
    else
        level = ctx->dbgcxfcn - 1;

    /* ask the debugger to do the enumeration */
    dbgenumlcl(ctx, level, cbfunc, cbctx);
}

/*
 *   Forget the current line.  This should be called just prior to
 *   resuming execution to ensure that the current line doesn't hang
 *   around while the game is executing for an extended period.  
 */
void CHtmlDebugHelper::forget_current_line()
{
    /* forget the current line if we have one */
    if (cur_is_valid_)
    {
        /* drop its current-line status flag */
        clear_line_status(cur_source_id_, cur_linenum_, HTMLTDB_STAT_CURLINE);

        /* forget we have a current line */
        cur_is_valid_ = FALSE;
    }

    /* if we have a context line, drop it as well */
    forget_ctx_line();
}

/*
 *   forget the context line 
 */
void CHtmlDebugHelper::forget_ctx_line()
{
    if (ctx_is_valid_)
    {
        /* drop the context-line status flag from the source line */
        clear_line_status(ctx_source_id_, ctx_linenum_, HTMLTDB_STAT_CTXLINE);

        /* forget about it */
        ctx_is_valid_ = FALSE;
    }

    /* clear the status flag in the stack window */
    if (ctx_stackwin_disp_ != 0 && stack_win_ != 0)
    {
        /* remove the flag */
        ctx_stackwin_disp_->set_dbg_status(stack_win_->win_, 0);
        
        /* forget the display item */
        ctx_stackwin_disp_ = 0;
    }
}

/*
 *   Update the debugger for a change in the execution status.  This
 *   should be called whenever we re-enter the debugger after executing
 *   game code.  We'll show the current line, reset the evaluation
 *   context, and update the stack window.  
 */
CHtmlSysWin *CHtmlDebugHelper::
   update_for_entry(dbgcxdef *ctx, CHtmlDebugSysIfc_win *syswinifc)
{
    CHtmlSysWin *win;

    /* if we have no context, there's nothing to do */
    if (ctx == 0)
        return 0;

    /* load the current source line */
    win = load_current_source(ctx, syswinifc);

    /* update the stack window */
    update_stack_window(ctx, TRUE);

    /* update the call history window, if we're keeping call history */
    if ((ctx->dbgcxflg & DBGCXFTRC) != 0)
        update_hist_window(ctx);

    /* return the current source line window */
    return win;
}

/*
 *   Load the source file at a particular stack level.  Opens the window,
 *   scrolls to the current line, and sets the appropriate line status at
 *   that line.  
 */
CHtmlSysWin *CHtmlDebugHelper::
   load_source_at_level(dbgcxdef *ctx, CHtmlDebugSysIfc_win *syswinifc,
                        int level)
{
    uchar infobuf[128];
    unsigned long linenum;
    CHtmlDbg_win_link *link;
    CHtmlDbg_src *src;

    /* if we previously had a current line showing on-screen, remove it */
    if (level == 0)
    {
        /* forget the current and context lines */
        forget_current_line();
    }
    else
    {
        /* forget only the old context line */
        forget_ctx_line();
    }

    /* if we have no context, we can't do anything here */
    if (ctx == 0)
        return 0;

    /* get information on the current line */
    if (dbglgetlvl(ctx, infobuf, level))
        return 0;

    /* search for a line source with the given ID */
    src = find_internal_src((int)infobuf[0]);

    /* if we didn't find a line source with the current line, give up */
    if (src == 0)
        return 0;

    /* load or find the source window for this line source */
    link = load_source_file(syswinifc, src);

    /* get the current line number */
    linenum = osrp4(infobuf + 1);

    /* go to the current line if we got a window */
    if (link != 0 && link->win_ != 0)
        show_source_line(link, linenum);

    /* set the current or the context line, as appropriate */
    if (level == 0)
    {
        /* set the new current line */
        cur_is_valid_ = TRUE;
        cur_source_id_ = src->source_id_;
        cur_linenum_ = linenum;

        /* update the status */
        set_line_status(cur_source_id_, cur_linenum_, HTMLTDB_STAT_CURLINE);
    }
    else
    {
        /* set the new context line */
        ctx_is_valid_ = TRUE;
        ctx_source_id_ = src->source_id_;
        ctx_linenum_ = linenum;

        /* update the status */
        set_line_status(ctx_source_id_, ctx_linenum_, HTMLTDB_STAT_CTXLINE);
    }

    /* return the source window we found */
    return link != 0 ? link->win_ : 0;
}

/*
 *   Enumerate open windows 
 */
void CHtmlDebugHelper::
   enum_source_windows(void (*func)(void *ctx, int idx,
                                    class CHtmlSysWin *win,
                                    int line_source_id,
                                    HtmlDbg_wintype_t win_type),
                       void *cbctx)
{
    CHtmlDbg_win_link *link;
    int idx;
    CHtmlDbg_win_link *next_link;

    /* go through the list of windows */
    for (link = first_srcwin_, idx = 0 ; link != 0 ; link = next_link, ++idx)
    {
        /* 
         *   remember the next window in the list - the callback could
         *   destroy this window before it returns, in which case we'd
         *   lose the window structure 
         */
        next_link = link->next_;
        
        /* invoke the callback for this window */
        (*func)(cbctx, idx, link->win_, link->source_id_, link->win_type_);
    }
}

/*
 *   Get a source window by index 
 */
CHtmlSysWin *CHtmlDebugHelper::get_source_window_by_index(int idx) const
{
    CHtmlDbg_win_link *link;

    /* go through the list of windows */
    for (link = first_srcwin_ ; link != 0 && idx != 0 ;
         link = link->next_, --idx) ;

    /* return the window we found */
    return (link == 0 ? 0 : link->win_);
}

/*
 *   Open a window on a line source, given the line source ID 
 */
CHtmlSysWin *CHtmlDebugHelper::
   open_line_source(CHtmlDebugSysIfc_win *syswinifc,
                    int line_source_id)
{
    CHtmlDbg_src *src;
    CHtmlDbg_win_link *link;

    /* find the line source */
    src = find_internal_src(line_source_id);

    /* if we didn't find it, we can't open a window */
    if (src == 0)
        return 0;
            
    /* open it */
    link = load_source_file(syswinifc, src);

    /* return the window */
    return (link == 0 ? 0 : link->win_);
}

/*
 *   find the window for a text file
 */
CHtmlSysWin *CHtmlDebugHelper::
   find_text_file(CHtmlDebugSysIfc_win *syswinifc,
                  const char *fname, const char *path)
{
    CHtmlDbg_src *src;
    CHtmlDbg_win_link *link;
    
    /* find the line source */
    src = find_internal_src(syswinifc, fname, path, FALSE);

    /* if we didn't find a line source, there's definitely no window */
    if (src == 0)
        return 0;

    /* search the open file list for the line source */
    for (link = first_srcwin_ ; link != 0 ; link = link->next_)
    {
        /* if this is the one, return it */
        if (syswinifc->dbgsys_fnames_eq(path, link->path_.get()))
            return link->win_;
    }

    /* we didn't find it */
    return 0;
}

/*
 *   Load an arbitrary text file into a window.
 */
CHtmlSysWin *CHtmlDebugHelper::
   open_text_file_at_line(CHtmlDebugSysIfc_win *syswinifc,
                          const char *fname, const char *path,
                          unsigned long linenum)
{
    CHtmlDbg_src *src;
    CHtmlDbg_win_link *link;

    /*
     *   Scan the source file line sources to see if we're actually
     *   opening one of the game's source files.  If so, just open the
     *   line source directly. 
     */
    src = find_internal_src(syswinifc, fname, path, TRUE);

    /* open it */
//$$$    link = load_source_file(syswinifc, src);
    link = load_source_file(syswinifc, fname, path, src->source_id_);

    /* 
     *   if we got a link, and we want to show a line other than the first
     *   line, go to the line number 
     */
    if (link != 0 && link->win_ != 0 && linenum != 1)
        show_source_line(link, linenum);
        
    /* return the line source's window */
    return (link == 0 ? 0 : link->win_);
}

/*
 *   Load a line source into a window 
 */
CHtmlDbg_win_link *CHtmlDebugHelper::
   load_source_file(CHtmlDebugSysIfc_win *syswinifc, CHtmlDbg_src *src)
{
    char path[OSFNMAX];
    char *rootname;

    /*
     *   Ignore the path in the source tracker and ask the UI to help us
     *   find the file.  In the past, we tried to be clever and use the
     *   stored path, but this can create problems when moving a
     *   configuration between directories or copying it to another
     *   machine, so overall it's better to use a simple approach and just
     *   search the UI's source path for the filename.  
     */
    rootname = os_get_root_name((char *)src->fname_.get());
    if (!dbgu_find_src(rootname, strlen(rootname), path, sizeof(path), TRUE))
    {
        /* the UI couldn't find it; give up */
        return 0;
    }

    /* save the path */
    src->path_.set(path);

    /* load the line source */
    return load_source_file(syswinifc, src->fname_.get(),
                            src->path_.get(), src->source_id_);
}

/*
 *   Load a source file 
 */
CHtmlDbg_win_link *CHtmlDebugHelper::
   load_source_file(CHtmlDebugSysIfc_win *syswinifc,
                    const textchar_t *fname, const textchar_t *path,
                    int source_id)
{
    CHtmlDbg_win_link *link;

    /*
     *   Search for the file in our open file list.  If we find it, simply
     *   return the existing window.  
     */
    for (link = first_srcwin_ ; link != 0 ; link = link->next_)
    {
        /* if this window's path matches the one we want, stop looking */
        if (syswinifc->dbgsys_fnames_eq(path, link->path_.get()))
            return link;
    }

    /* 
     *   we didn't find it in our open file list, so we'll have to load it
     *   for real - create a new window, and load the file into the new
     *   window 
     */
    link = create_source_win(syswinifc, fname, path, source_id);
    load_file_into_win(link, path, TRUE);

    /* return the new window */
    return link;
}
    
/*
 *   Load a source file into a given window 
 */
int CHtmlDebugHelper::load_file_into_win(CHtmlDbg_win_link *win_link,
                                         const textchar_t *fname,
                                         int format)
{
    osfildef *fp;
    struct
    {
        linfdef linf;
        char fname[256];
    } linf;
    lindef *lin;
    uchar posbuf[4];
    CHtmlRect rc;
    ulong linenum;

    /* get the base class line source */
    lin = &linf.linf.linflin;

    /* open the file */
    fp = osfoprs(fname, OSFTTEXT);

    /* if we couldn't open the file, return failure */
    if (fp == 0)
        return 1;

    /* set up a line source to read the file */
    linfini2(0, &linf.linf, (char *)fname, get_strlen(fname), fp, TRUE);
    linf.linf.linflin.linid = win_link->source_id_;

    /* seek to the start fo the line source */
    oswp4(posbuf, 0);
    linseek(lin, posbuf);

    /* make sure there's nothing on the page */
    win_link->parser_->clear_page();

    /* 
     *   read lines from the line source, and add each line to the
     *   formatter list for the window 
     */
    for (linenum = 0 ; ; )
    {
        textchar_t buf[1024];
        CHtmlTagText *txt_tag;
        textchar_t *p;

        /* get the next line */
        if (!lingets(lin, (uchar *)buf, sizeof(buf)))
            break;

        /* count the line */
        ++linenum;

        /* remove any newline characters at the end of the line */
        for (p = buf + get_strlen(buf) ; p > buf ; )
        {
            /* move to the previous character */
            --p;

            /* if it's not a \r or \n, we're done */
            if (*p != '\n' && *p != '\r')
                break;

            /* remove it and continue scanning */
            *p = '\0';
        }

        /* add a text tag for the line source */
        txt_tag =
            new CHtmlTagTextDebugsrc(linenum, tab_size_,
                                     win_link->parser_->get_text_array(),
                                     buf, get_strlen(buf));
        win_link->parser_->append_tag(txt_tag);
    }

    /* 
     *   Add an extra blank line at the end, to ensure that the final line
     *   gets formatted properly.  The formatter gets confused without a
     *   final empty line.  
     */
    win_link->parser_->append_tag(
        new CHtmlTagTextDebugsrc(linenum + 1, tab_size_,
                                 win_link->parser_->get_text_array(), "", 0));

    /* if desired, format the text */
    if (format)
    {
        /* do the initial formatting of the window */
        win_link->win_->do_formatting(TRUE, TRUE, TRUE);

        /* update the scrollbars */
        win_link->win_->fmt_adjust_vscroll();

        /* update special-line indicators for this window */
        update_lines_on_screen(win_link);

        /* scroll the window to the top */
        rc.set(0, 0, 0, 0);
        win_link->win_->scroll_to_doc_coords(&rc);
    }

    /* done with the source file - close it */
    osfcls(fp);

    /* success */
    return 0;
}

/*
 *   Go through our special-line records and update the display so that a
 *   window's settings reflect the current special-line status flags.
 *   Note that we only have to do this for line sources with valid (i.e.,
 *   other than -1) id's, since these are the only sources that can have
 *   special line status flags.  
 */
void CHtmlDebugHelper::update_lines_on_screen(CHtmlDbg_win_link *win_link)
{
    CHtmlDbg_line_link *line_link;

    if (win_link->source_id_ != -1)
    {
        /* look through our list of special lines */
        for (line_link = first_line_ ; line_link != 0 ;
             line_link = line_link->next_)
        {
            /* if this line is from the source we just loaded, update it */
            if (line_link->source_id_ == win_link->source_id_)
            {
                /* update this line on-screen */
                update_line_on_screen(win_link, line_link);
            }
        }
    }
}


/*
 *   Create a new window for displaying a source file
 */
CHtmlDbg_win_link *CHtmlDebugHelper::create_source_win(
    CHtmlDebugSysIfc_win *syswinifc, const textchar_t *fname,
    const textchar_t *path, int source_id)
{
    /* create the window */
    return create_win(syswinifc, fname, fname, path, source_id,
                      HTMLDBG_WINTYPE_SRCFILE);
}

/*
 *   Set the tab expansion size to use in source windows.  If this changes
 *   while windows are open, we'll need to reload files into all open
 *   windows, since we translate tabs while loading files. 
 */
void CHtmlDebugHelper::set_tab_size(class CHtmlDebugSysIfc_win *syswinifc,
                                    int n)
{
    if (tab_size_ != n && tab_size_ != 0)
    {
        /* remember the new tab size */
        tab_size_ = n;
        
        /* reload all source windows */
        reload_source_windows(syswinifc);
    }
}

/*
 *   Reformat all source windows 
 */
void CHtmlDebugHelper::
   reformat_source_windows(dbgcxdef *ctx, CHtmlDebugSysIfc_win *syswinifc)
{
    CHtmlDbg_win_link *link;

    /* go through the list of windows, and reformat each */
    for (link = first_srcwin_ ; link != 0 ; link = link->next_)
        reformat_source_window(ctx, syswinifc, link, 0);
}

/*
 *   Reformat a source file 
 */
void CHtmlDebugHelper::
   reformat_source_window(dbgcxdef *ctx,
                          CHtmlDebugSysIfc_win *syswinifc,
                          CHtmlDbg_win_link *link,
                          const CHtmlRect *orig_scroll_pos)
{
    CHtmlFontDesc font_desc;
    CHtmlRect rc;

    /* 
     *   If the caller passed in the original scrolling position, use that
     *   position; otherwise, save the current scrolling position.  
     */
    if (orig_scroll_pos != 0)
        rc = *orig_scroll_pos;
    else
        link->win_->get_scroll_doc_coords(&rc);
        
    /* reset the window */
    link->formatter_->start_at_top(TRUE);

    /* get the font, in case it's changed */
    font_desc.fixed_pitch = TRUE;
    font_desc.htmlsize = 3;
    syswinifc->dbgsys_set_srcfont(&font_desc);
    link->font_ = link->win_->get_font(&font_desc);
    link->formatter_->set_font(link->font_);

    /* update the window according to what type of window we have */
    switch(link->win_type_)
    {
    case HTMLDBG_WINTYPE_SRCFILE:
    case HTMLDBG_WINTYPE_HISTORY:
    case HTMLDBG_WINTYPE_DEBUGLOG:
        /* do a complete reformat of this window */
        link->win_->do_formatting(TRUE, TRUE, TRUE);
        link->win_->fmt_adjust_vscroll();

        /* make sure we've updated line indicators */
        update_lines_on_screen(link);

        /* scroll back to the original position in the window */
        link->win_->scroll_to_doc_coords(&rc);

        /* done */
        break;

    case HTMLDBG_WINTYPE_STACK:
        /* 
         *   for the stack window, just rebuild it from scratch - it's
         *   easier than duplicating the code that figures out where all
         *   the special indicators go 
         */
        update_stack_window(ctx, FALSE);
        break;
    }
}

/*
 *   Reload a single source window.  If win == 0, reloads all source
 *   windows.  
 */
void CHtmlDebugHelper::reload_source_window(CHtmlDebugSysIfc_win *syswinifc,
                                            CHtmlSysWin *win)
{
    CHtmlDbg_win_link *link;

    /* go through the list of windows */
    for (link = first_srcwin_ ; link != 0 ; link = link->next_)
    {
        /* 
         *   if it's a source file, and either it matches the target
         *   window or we're reloading all windows (win == 0), reload this
         *   window 
         */
        if (link->win_type_ == HTMLDBG_WINTYPE_SRCFILE
            && (win == link->win_ || win == 0))
        {
            const textchar_t *path;
            CHtmlRect orig_scroll_pos;

            /* 
             *   open the file - use the full path name if present,
             *   otherwise use the filename 
             */
            if (link->path_.get() != 0 && link->path_.get()[0] != '\0')
                path = link->path_.get();
            else
                path = link->fname_.get();

            /* make sure we can access the file */
            if (osfacc(path))
                continue;

            /* 
             *   before re-loading the file, remember the original
             *   scrolling position in the window, so that we can restore
             *   the original position after we finish reloading the file 
             */
            link->win_->get_scroll_doc_coords(&orig_scroll_pos);

            /* clear the page */
            link->formatter_->start_at_top(FALSE);

            /* load it */
            load_file_into_win(link, path, FALSE);

            /* reformat it */
            reformat_source_window(0, syswinifc, link, &orig_scroll_pos);

            /* 
             *   if we're reloading only one window, there's no need to
             *   look any further 
             */
            if (win != 0)
                break;
        }
        else
        {
            /* 
             *   if we're doing all windows, reformat this one even though
             *   it's not a source window 
             */
            if (win == 0)
                reformat_source_window(0, syswinifc, link, 0);
        }
    }
}

/*
 *   get the stack window 
 */
CHtmlSysWin *CHtmlDebugHelper::get_stack_win() const
{
    return (stack_win_ != 0 ? stack_win_->win_ : 0);
}

/*
 *   get the history window 
 */
CHtmlSysWin *CHtmlDebugHelper::get_hist_win() const
{
    return (hist_win_ != 0 ? hist_win_->win_ : 0);
}

/*
 *   get the debug log window 
 */
CHtmlSysWin *CHtmlDebugHelper::get_debuglog_win() const
{
    return (debuglog_win_ != 0 ? debuglog_win_->win_ : 0);
}

/*
 *   Create a window to use for stack display 
 */
CHtmlSysWin *CHtmlDebugHelper::
   create_stack_win(CHtmlDebugSysIfc_win *syswinifc)
{
    /* create a new stack window if we don't already have one */
    if (stack_win_ == 0)
        stack_win_ = create_win(syswinifc, "Stack", 0, 0, -1,
                                HTMLDBG_WINTYPE_STACK);

    /* return the system window */
    return stack_win_->win_;
}

/*
 *   Create the debug log window 
 */
CHtmlSysWin *CHtmlDebugHelper::
   create_debuglog_win(CHtmlDebugSysIfc_win *syswinifc)
{
    /* create a new debug log window if we don't have one yet */
    if (debuglog_win_ == 0)
        debuglog_win_ = create_win(syswinifc, "Debug Log", 0, 0, -1,
                                   HTMLDBG_WINTYPE_DEBUGLOG);

    /* return the system window */
    return debuglog_win_->win_;
}

/*
 *   Create a window to use for history display 
 */
CHtmlSysWin *CHtmlDebugHelper::
   create_hist_win(CHtmlDebugSysIfc_win *syswinifc)
{
    /* create a new stack window if we don't already have one */
    if (hist_win_ == 0)
        hist_win_ = create_win(syswinifc, "Call History", 0, 0, -1,
                               HTMLDBG_WINTYPE_HISTORY);

    /* return the system window */
    return hist_win_->win_;
}


/*
 *   Create a debugger window.  linf can be null (this will be the case
 *   for system windows, such as the stack window, not tied to a line
 *   source).  
 */
CHtmlDbg_win_link *CHtmlDebugHelper::
   create_win(CHtmlDebugSysIfc_win *syswinifc,
              const textchar_t *win_title,
              const textchar_t *fname, const textchar_t *path,
              int source_id, HtmlDbg_wintype_t win_type)
{
    CHtmlParser *parser;
    CHtmlFormatter *formatter;
    CHtmlSysWin *win;
    CHtmlFontDesc font_desc;
    CHtmlSysFont *font;

    /* 
     *   if there's an underlying file, make sure it exists - if not, fail
     *   to create the window 
     */
    if (path != 0 && osfacc(path))
        return 0;

    /* create a parser for the window */
    parser = new CHtmlParser(TRUE);

    /* create a formatter for the window */
    formatter = new CHtmlFormatterMain(parser);

    /* if there's no path, use the filename instead */
    if (path == 0)
        path = fname;

    /* create the system window using the system interface */
    win = syswinifc
          ->dbgsys_create_win(parser, formatter, win_title, path,
                              win_type);

    /* set up the formatter with the default monospaced font */
    font_desc.fixed_pitch = TRUE;
    font_desc.htmlsize = 3;
    syswinifc->dbgsys_set_srcfont(&font_desc);
    font = win->get_font(&font_desc);
    formatter->set_font(font);

    /* 
     *   add the window to our list of source windows, and return the new
     *   link structure 
     */
    return add_src_win(win, parser, formatter,  win_title,
                       fname, path, source_id, font, win_type);
}

/*
 *   Add a window to the source window list 
 */
CHtmlDbg_win_link *CHtmlDebugHelper::add_src_win(
    CHtmlSysWin *win, CHtmlParser *parser, CHtmlFormatter *formatter,
    const textchar_t *win_title, const textchar_t *fname,
    const textchar_t *path, int source_id, CHtmlSysFont *font,
    HtmlDbg_wintype_t win_type)
{
    CHtmlDbg_win_link *link;
    
    /* create the new link */
    link = new CHtmlDbg_win_link(win, parser, formatter, win_title,
                                 fname, path, source_id, font, win_type);

    /* link it in at the end of our list of source windows */
    if (last_srcwin_ != 0)
        last_srcwin_->next_ = link;
    else
        first_srcwin_ = link;
    link->prev_ = last_srcwin_;
    last_srcwin_ = link;

    /* return the new link */
    return link;
}

/*
 *   Delete a window link from the source window list.  Removes the link
 *   from the list and deletes the link object.  
 */
void CHtmlDebugHelper::del_src_win(CHtmlDbg_win_link *link)
{
    /* unlink it from its previous item */
    if (link->prev_ != 0)
        link->prev_->next_ = link->next_;
    else
        first_srcwin_ = link->next_;

    /* unlink it from its next item */
    if (link->next_ != 0)
        link->next_->prev_ = link->prev_;
    else
        last_srcwin_ = link->prev_;

    /* delete the link */
    delete link;
}

/*
 *   Receive notification that a source window is being closed 
 */
void CHtmlDebugHelper::on_close_srcwin(CHtmlSysWin *win)
{
    CHtmlDbg_win_link *link;
    
    /* find the window in our list */
    for (link = first_srcwin_ ; link != 0 ; link = link->next_)
    {
        /* if this is the one, delete it */
        if (link->win_ == win)
        {
            /* this is it - delete the formatter and parser */
            link->formatter_->release_parser();
            delete link->parser_;
            delete link->formatter_;

            /* if this is the stack window, forget that as well */
            if (link == stack_win_)
                stack_win_ = 0;

            /* if it's the debug log window, forget it */
            if (link == debuglog_win_)
                debuglog_win_ = 0;

            /* if it's the history window, forget it, too */
            if (link == hist_win_)
                hist_win_ = 0;

            /* now delete the link itself */
            del_src_win(link);

            /* we're done */
            return;
        }
    }

    /* we didn't find it; ignore the request */
}

/*
 *   Get the line and window link records for the current selection in a
 *   given window.  Returns the file seek position of the start of the
 *   line containing the start of the selection.  
 */
unsigned long CHtmlDebugHelper::
   find_line_info(CHtmlSysWin *win, CHtmlDbg_win_link **win_linkp,
                  CHtmlDbg_line_link **line_linkp) const
{
    CHtmlDbg_win_link *wlink;
    
    /* presume we won't find anything */
    *win_linkp = 0;
    *line_linkp = 0;

    /* find the window in our list of source windows */
    for (wlink = first_srcwin_ ; wlink != 0 ; wlink = wlink->next_)
    {
        /* if this is the one, find the current selection in it */
        if (wlink->win_ == win)
        {
            unsigned long sel_start, sel_end;
            unsigned long linenum;
            CHtmlDispTextDebugsrc *disp;

            /* get the start of the current selection */
            wlink->formatter_->get_sel_range(&sel_start, &sel_end);

            /* find the display item at the start of the selection */
            disp = (CHtmlDispTextDebugsrc *)
                   wlink->formatter_->find_by_txtofs(sel_start, TRUE, FALSE);
            if (disp == 0)
                return 0;

            /* get the file position of the display item */
            linenum = disp->get_debugsrcpos();

            /* tell the caller about the window */
            *win_linkp = wlink;

            /* find the line record */
            *line_linkp = find_line_link(wlink->source_id_, linenum);

            /* return the file position */
            return linenum;
        }
    }

    /* didn't find anything */
    return 0;
}


/*
 *   Get the line link record for a given source line 
 */
CHtmlDbg_line_link *CHtmlDebugHelper::find_line_link(
    int line_source_id, unsigned long linenum) const
{
    CHtmlDbg_line_link *link;

    /* find the line in our list */
    for (link = first_line_ ; link != 0 ; link = link->next_)
    {
        /* see if this is the one we're looking for */
        if (link->equals(line_source_id, linenum))
        {
            /* this is it - return it */
            return link;
        }
    }

    /* didn't find it */
    return 0;
}


/*
 *   Get information on the source line containing the given breakpoint 
 */
int CHtmlDebugHelper::get_bp_line_info(int bpnum, int *source_id,
                                       unsigned long *linenum)
{
    CHtmlDbg_bp *bp;
    
    /* find the breakpoint */
    bp = find_internal_bp(bpnum);
    if (bp == 0)
        return FALSE;

    /* check the breakpoint type */
    if (bp->global_)
    {
        /* there's no line information for a global breakpoint */
        *source_id = -1;
        *linenum = 0;
        return FALSE;
    }
    else
    {
        /* it's a source line - return the information */
        *source_id = bp->source_id_;
        *linenum = bp->linenum_;
        return TRUE;
    }
}

/*
 *   Set or clear a source line status bit
 */
void CHtmlDebugHelper::change_line_status(int line_source_id,
                                          unsigned long seekpos,
                                          unsigned int statflag, int set)
{
    CHtmlDbg_line_link *link;

    /* find the link */
    link = find_line_link(line_source_id, seekpos);

    /* if we found it, adjust the current flags as requested */
    if (link != 0)
    {
        /* set or clear the new flags as needed */
        if (set)
            link->stat_ |= statflag;
        else
            link->stat_ &= ~statflag;
        
        /* update the line on-screen */
        update_line_on_screen(link);
        
        /* 
         *   if the flags are now all cleared in this line, drop the line
         *   from the special lines list, since there's nothing special to
         *   track about it any more 
         */
        if (link->stat_ == 0)
        {
            /* unlink it from the list */
            if (link->prev_ == 0)
                first_line_ = link->next_;
            else
                link->prev_->next_ = link->next_;
            if (link->next_ == 0)
                last_line_ = link->prev_;
            else
                link->next_->prev_ = link->prev_;
            
            /* delete it */
            delete link;
        }
    }
    else
    {
        /*
         *   We didn't find the line we were looking for, so the line
         *   doesn't currently have any special properties.  If we're
         *   clearing bits, there's nothing to do, since the line already
         *   has all its status bits cleared.  
         */
        if (!set)
            return;
        
        /* create a new special-line link and add it to the list */
        link = new CHtmlDbg_line_link(line_source_id, seekpos, statflag);
        link->prev_ = last_line_;
        link->next_ = 0;
        if (last_line_ != 0)
            last_line_->next_ = link;
        else
            first_line_ = link;
        last_line_ = link;
        
        /* update the line on-screen */
        update_line_on_screen(link);
    }
}

/*
 *   Update a line on-screen 
 */
void CHtmlDebugHelper::update_line_on_screen(CHtmlDbg_line_link *link)
{
    CHtmlDbg_win_link *win;
    
    /*
     *   Find the line's file in our file list.  If the line isn't in our
     *   file list anywhere, it's not displayed anywhere on screen, so
     *   there's nothing to update just now. 
     */
    for (win = first_srcwin_ ; win != 0 ; win = win->next_)
    {
        /* 
         *   See if this window's line source ID matches the line's source
         *   ID.  Note that we'll keep going even if we find a match,
         *   since we could conceivably have the same file open in
         *   multiple windows.  
         */
        if (win->source_id_ == link->source_id_)
            update_line_on_screen(win, link);
    }
}

/*
 *   update a particular line in a particular window on-screen 
 */
void CHtmlDebugHelper::update_line_on_screen(CHtmlDbg_win_link *win,
                                             CHtmlDbg_line_link *link)
{
    CHtmlDispTextDebugsrc *disp;

    /* find the display item for this line */
    disp = (CHtmlDispTextDebugsrc *)win->formatter_->
           find_by_debugsrcpos(link->linenum_);

    /* if we found it, update its status */
    if (disp != 0)
        disp->set_dbg_status(win->win_, link->stat_);
}


/*
 *   Context structure for the stack iterator callback 
 */
struct HtmlDbg_stkdisp_t
{
    HtmlDbg_stkdisp_t(CHtmlDebugHelper *helper)
    {
        /* remember the helper object */
        helper_ = helper;

        /* start out at level 0 */
        level_ = 0;
    }
    
    /* the helper object */
    CHtmlDebugHelper *helper_;

    /* stack level - we increment this for each line in the stack trace */
    int level_;

    /* buffer for batching up each line of text */
    CStringBuf buf_;
};

/*
 *   Update the stack window 
 */
void CHtmlDebugHelper::update_stack_window(dbgcxdef *ctx,
                                           int reset_level)
{
    HtmlDbg_stkdisp_t cbctx(this);
    static textchar_t emptymsg[] = "[stack trace unavailable]";
    CHtmlDispTextDebugsrc *disp;
    int level;
    
    /* if we don't have a stack window, there's nothing to update */
    if (stack_win_ == 0)
        return;

    /* delete everything currently in the stack window */
    stack_win_->formatter_->start_at_top(TRUE);
    stack_win_->formatter_->set_font(stack_win_->font_);
    stack_win_->parser_->clear_page();

    /* determine what level we'll be at when we're done */
    if (reset_level)
    {
        /* we're resetting the level to zero */
        level = 0;
    }
    else
    {
        /* remember the current level */
        level = (ctx_stackwin_disp_ == 0
                 ? 0 : ctx_stackwin_disp_->get_debugsrcpos());
    }
    
    /* 
     *   make sure we're not holding on to a context display item, since
     *   we're about to delete all of the display items in the stack
     *   window 
     */
    ctx_stackwin_disp_ = 0;
    
    /* build the stack */
    if (ctx != 0)
        dbgstktr(ctx, &update_stack_disp_cb, &cbctx, 0, FALSE, FALSE);

    /*
     *   If there's nothing on the stack at all, it means the game is
     *   running.  Put an indication of this in the stack listing so that
     *   it's not completely empty. 
     */
    if (cbctx.level_ == 0)
        stack_win_->parser_->append_tag(new CHtmlTagTextDebugsrc(
            0, tab_size_, stack_win_->parser_->get_text_array(),
            emptymsg, get_strlen(emptymsg)));

    /* add a blank line at the end */
    stack_win_->parser_->append_tag(new CHtmlTagTextDebugsrc(
        0, tab_size_, stack_win_->parser_->get_text_array(), "", 0));

    /* format the stack into the stack window */
    if (!stack_win_->win_->do_formatting(TRUE, TRUE, TRUE))
    {
        CHtmlRect area(0, 0, HTMLSYSWIN_MAX_RIGHT, HTMLSYSWIN_MAX_BOTTOM);
        stack_win_->win_->inval_doc_coords(&area);
    }

    /* update the scrollbars */
    stack_win_->win_->fmt_adjust_vscroll();

    /* 
     *   Turn on the current-line status flag for the first item - the
     *   first item in the list is always the current line, so display it
     *   thus.  Don't do this if there's no stack traceback available.  
     */
    disp = (CHtmlDispTextDebugsrc *)
           stack_win_->formatter_->find_by_debugsrcpos(0);
    if (disp != 0)
        disp->set_dbg_status(stack_win_->win_,
                             cbctx.level_ == 0 ? 0 :
                             HTMLTDB_STAT_CURLINE);

    /* set the context line, if we're at a nested context */
    if (level != 0)
    {
        /* get the correct level */
        ctx_stackwin_disp_ = (CHtmlDispTextDebugsrc *)
            stack_win_->formatter_->find_by_debugsrcpos(level);

        /* set its status */
        if (ctx_stackwin_disp_ != 0)
            ctx_stackwin_disp_->set_dbg_status(stack_win_->win_,
                                               HTMLTDB_STAT_CTXLINE);
    }
}

/*
 *   Display callback for for update_stack_window 
 */
void CHtmlDebugHelper::update_stack_disp_cb(void *ctx0, char *str, int strl)
{
    HtmlDbg_stkdisp_t *ctx = (HtmlDbg_stkdisp_t *)ctx0;
    
    /* invoke the non-static callback handler */
    ctx->helper_->update_stack_disp(ctx, str, strl);
}

void CHtmlDebugHelper::update_stack_disp(HtmlDbg_stkdisp_t *ctx,
                                         char *str, int strl)
{
    char *p;
    const char *p_end;

    /* look for a newline */
    for (p = str, p_end = str + strl ; p < p_end ; ++p)
    {
        /* if this is a newline, add the line to the window */
        if (*p == '\n')
        {
            CHtmlTag *tag;

            /* if we have anything prior to the newline, add to the window */
            if (p != str)
            {
                /* append up to the newline to our line buffer */
                ctx->buf_.append(str, p - str);
            }

            /* 
             *   Add the line in the buffer to our list.  Note that we use
             *   the line number (i.e., the stack nesting level) as the
             *   position. 
             */
            tag = new CHtmlTagTextDebugsrc
                  (ctx->level_, tab_size_,
                   stack_win_->parser_->get_text_array(),
                   ctx->buf_.get(), get_strlen(ctx->buf_.get()));
            stack_win_->parser_->append_tag(tag);

            /* count the line */
            ctx->level_++;

            /* the buffer has been used, so clear it */
            ctx->buf_.set("");
            
            /* drop everything prior to the newline */
            str = p + 1;
        }
    }
    
    /* add the remainder of the line to the buffer */
    if (p > str)
        ctx->buf_.append(str, p - str);
}

/*
 *   Activate a stack window item 
 */
CHtmlSysWin *CHtmlDebugHelper::
   activate_stack_win_item(CHtmlDisp *disp, dbgcxdef *ctx,
                           CHtmlDebugSysIfc_win *syswinifc)
{
    CHtmlDispTextDebugsrc *dbgdisp = (CHtmlDispTextDebugsrc *)disp;
    unsigned long level;

    /* if there's no context, there's nothing to do */
    if (ctx == 0)
        return 0;
    
    /* forget any existing context line we have */
    forget_ctx_line();

    /* 
     *   get the level from the item - we store this as the file position
     *   of the item 
     */
    level = dbgdisp->get_debugsrcpos();

    /* activate the level */
    return activate_stack_win_line(level, ctx, syswinifc);
}

/*
 *   Activate a stack window line 
 */
CHtmlSysWin *CHtmlDebugHelper::
   activate_stack_win_line(int level, dbgcxdef *ctx,
                           CHtmlDebugSysIfc_win *syswinifc)
{
    /* 
     *   if it's level 0, select the current line; otherwise, activate the
     *   enclosing stack context level 
     */
    if (level == 0)
    {
        /* no new context line - activate the current line */
        return load_current_source(ctx, syswinifc);
    }
    else
    {
        CHtmlSysWin *win;
        CHtmlDispTextDebugsrc *dbgdisp;

        /* load the source at this level */
        win = load_source_at_level(ctx, syswinifc, level);

        /* update the stack window context display if possible */
        if (stack_win_ != 0)
        {
            /* find the display item for this line */
            dbgdisp = (CHtmlDispTextDebugsrc *)
                      stack_win_->formatter_->get_disp_by_linenum(level);
            
            /* set the stack window to show that the new context line */
            dbgdisp->set_dbg_status(stack_win_->win_, HTMLTDB_STAT_CTXLINE);
            ctx_stackwin_disp_ = dbgdisp;
        }
            
        /* return the window */
        return win;
    }
}

/*
 *   Update the history window 
 */
void CHtmlDebugHelper::update_hist_window(dbgcxdef *ctx)
{
    const textchar_t *startp;
    const textchar_t *endp;
    const textchar_t *p;

    /* if there's no context, ignore it */
    if (ctx == 0)
        return;

    /* if there's no history window, there's nothing to update */
    if (hist_win_ == 0)
        return;

    /* delete everything currently in the window */
    hist_win_->formatter_->start_at_top(TRUE);
    hist_win_->formatter_->set_font(hist_win_->font_);
    hist_win_->parser_->clear_page();
    
    /* run through the history buffer */
    startp = get_call_trace_buf(ctx);
    endp = startp + get_call_trace_len(ctx);
    for (p = startp ; p < endp ; p += get_strlen(p) + 1)
    {
        CHtmlTag *tag;
        
        /* store this line */
        tag = new CHtmlTagTextDebugsrc
              (0, tab_size_, hist_win_->parser_->get_text_array(),
               p, get_strlen(p));
        hist_win_->parser_->append_tag(tag);
    }

    /* add a blank line at the end */
    hist_win_->parser_->append_tag(new CHtmlTagTextDebugsrc(
        0, tab_size_, hist_win_->parser_->get_text_array(), "", 0));

    /* format it into the window */
    if (!hist_win_->win_->do_formatting(TRUE, TRUE, TRUE))
    {
        CHtmlRect area(0, 0, HTMLSYSWIN_MAX_RIGHT, HTMLSYSWIN_MAX_BOTTOM);
        hist_win_->win_->inval_doc_coords(&area);
    }

    /* update the scrollbars */
    hist_win_->win_->fmt_adjust_vscroll();
}

/*
 *   Toggle hidden output 
 */
void CHtmlDebugHelper::toggle_dbg_hid()
{
    dbghid = !dbghid;
}

/*
 *   Get hidden output display status 
 */
int CHtmlDebugHelper::get_dbg_hid()
{
    return dbghid;
}

/*
 *   Enumerate line sources 
 */
void CHtmlDebugHelper::
   enum_source_files(void (*func)(void *cbctx, const char *fname,
                                  int line_source_id), void *cbctx)
{
    CHtmlDbg_src *src;
    CHtmlDbg_src *nxt;
    
    /* run through the line sources, and invoke the callback for each one */
    for (src = src_ ; src != 0 ; src = nxt)
    {
        /* remember the next one, in case the callback deletes this one */
        nxt = src->nxt_;

        /* invoke the callback */
        (*func)(cbctx, src->fname_.get(), src->source_id_);
    }
}

/*
 *   Find an internal source tracker 
 */
CHtmlDbg_src *CHtmlDebugHelper::find_internal_src(int source_id) const
{
    CHtmlDbg_src *src;

    /* scan our list */
    for (src = src_ ; src != 0 ; src = src->nxt_)
    {
        /* if this one matches, return it */
        if (src->source_id_ == source_id)
            return src;
    }

    /* didn't find it */
    return 0;
}

/*
 *   Find an internal source tracker by filename 
 */
CHtmlDbg_src *CHtmlDebugHelper::
   find_internal_src(CHtmlDebugSysIfc_win *syswinifc,
                     const textchar_t *fname, const textchar_t *path,
                     int add_if_not_found)
{
    CHtmlDbg_src *src;

    /* 
     *   if the given filename is absolute, and there's no path, consider
     *   the filename to be the path 
     */
    if ((path == 0 || path[0] == '\0') && os_is_file_absolute(fname))
        path = fname;
    
    /* scan our list */
    for (src = src_ ; src != 0 ; src = src->nxt_)
    {
        int match;
        
        /* 
         *   Compare the filenames.  If we have a path and the source has
         *   a path, compare the paths.  Otherwise, compare just the
         *   display filenames.  
         */
        if (path == 0 || path[0] == '\0'
            || src->path_.get() == 0 || src->path_.get()[0] == '\0')
        {
            /* 
             *   if the file we're looking for has a path and the line
             *   source doesn't, check to see if the line source's
             *   filename matches the given path 
             */
            if (path != 0 && path[0] != '\0')
            {
                /* 
                 *   the file has a path, even though the line source
                 *   doesn't - compare the filename to see if it belongs
                 *   in the path 
                 */
                match = syswinifc
                        ->dbgsys_fname_eq_path(path, src->fname_.get());

                /* 
                 *   if we matched, set the line source's path to the full
                 *   path of this file, so that we look for this line
                 *   source in this location from now on 
                 */
                if (match)
                    src->path_.set(path);
            }
            else if (src->path_.get() != 0 && src->path_.get()[0] != '\0')
            {
                /* 
                 *   the line source has a path, even though the file
                 *   doesn't - compare the filename to see if it belongs
                 *   in the path 
                 */
                match = syswinifc
                        ->dbgsys_fname_eq_path(src->path_.get(), fname);
            }
            else
            {
                /* we don't have any paths - compare only the filenames */
                match = syswinifc
                        ->dbgsys_fnames_eq(fname, src->fname_.get());
            }
        }
        else
        {
            /* we have full paths - compare them */
            match = syswinifc->dbgsys_fnames_eq(path, src->path_.get());
        }

        /* if they match, this is the source we want */
        if (match)
            return src;
    }

    /* 
     *   we didn't find it - if they want to create a new one in this
     *   case, do so now 
     */
    if (add_if_not_found)
        return add_fake_internal_line_source(fname);

    /* not found */
    return 0;
}

/*
 *   Find an internal breakpoint tracker 
 */
CHtmlDbg_bp *CHtmlDebugHelper::find_internal_bp(int bpnum) const
{
    CHtmlDbg_bp *bp;

    /* scan our list */
    for (bp = bp_ ; bp != 0 ; bp = bp->nxt_)
    {
        /* if this is the one, return it */
        if (bp->bpnum_ == bpnum)
            return bp;
    }

    /* didn't find it */
    return 0;
}

/*
 *   Find an internal breakpoint tracker 
 */
CHtmlDbg_bp *CHtmlDebugHelper::find_internal_bp(int source_id,
                                                unsigned long linenum) const
{
    CHtmlDbg_bp *bp;

    /* scan our list */
    for (bp = bp_ ; bp != 0 ; bp = bp->nxt_)
    {
        /* if this is the one, return it */
        if (!bp->global_
            && bp->source_id_ == source_id
            && bp->linenum_ == linenum)
            return bp;
    }

    /* didn't find it */
    return 0;
}

/*
 *   delete a breakpoint tracker 
 */
void CHtmlDebugHelper::delete_internal_bp(CHtmlDbg_bp *bp_to_del)
{
    CHtmlDbg_bp *bp;
    CHtmlDbg_bp *prv;

    /* find the entry in the list */
    for (prv = 0, bp = bp_ ; bp != 0 ; prv = bp, bp = bp->nxt_)
    {
        /* if this is the one, delete it from the list */
        if (bp == bp_to_del)
        {
            /* unlink it from the list */
            if (prv == 0)
                bp_ = bp->nxt_;
            else
                prv->nxt_ = bp->nxt_;

            /* delete it */
            delete bp;

            /* done */
            return;
        }
    }
}

/*
 *   Synthesize a breakpoint ID 
 */
int CHtmlDebugHelper::synthesize_bp_num() const
{
    int max_bpnum;
    CHtmlDbg_bp *bp;
    
    /* find the maximum breakpoint number currently in use */
    for (bp = bp_, max_bpnum = 0 ; bp != 0 ; bp = bp->nxt_)
    {
        /* if this is higher than the highest we've seen yet, record it */
        if (bp->bpnum_ > max_bpnum)
            max_bpnum = bp->bpnum_;
    }
    
    /* 
     *   return one higher than the maximum currently in use - this is
     *   guaranteed to be unique, since it's not equal to anything in the
     *   list 
     */
    return max_bpnum + 1;
}


/* ------------------------------------------------------------------------ */
/*
 *   CHtmlDbg_win_link implementation 
 */

CHtmlDbg_win_link::CHtmlDbg_win_link(CHtmlSysWin *win, CHtmlParser *parser,
                                     CHtmlFormatter *formatter,
                                     const textchar_t *win_title,
                                     const textchar_t *fname,
                                     const textchar_t *path,
                                     int source_id,
                                     CHtmlSysFont *font,
                                     HtmlDbg_wintype_t win_type)
{
    /* remember window and parser */
    win_ = win;
    parser_ = parser;
    formatter_ = formatter;

    /* not in a list yet, so clear link pointers */
    next_ = prev_ = 0;

    /* remember the name of the file */
    fname_.set(fname != 0 ? fname : win_title);

    /* remember the path to the file */
    path_.set(path);

    /* remember the line source ID */
    source_id_ = source_id;

    /* remember the default font */
    font_ = font;

    /* remember the type of the window */
    win_type_ = win_type;
}

/*
 *   Find the line source for a given line source ID 
 */
lindef *CHtmlDbg_win_link::find_lindef(struct dbgcxdef *ctx, int id)
{
    lindef *lin;

    /* scan the list of line sources */
    for (lin = ctx->dbgcxlin ; lin != 0 ; lin = lin->linnxt)
    {
        /* if this is the one, return it */
        if (lin->linid == id)
            return lin;
    }

    /* not found */
    return 0;
}

/* ------------------------------------------------------------------------ */
/*
 *   Special-line record 
 */

CHtmlDbg_line_link::CHtmlDbg_line_link(int source_id, unsigned long linenum,
                                       unsigned int init_stat)
{
    /* remember the information */
    source_id_ = source_id;
    linenum_ = linenum;
    stat_ = init_stat;

    /* not in a list yet */
    next_ = prev_ = 0;
}

/* ------------------------------------------------------------------------ */
/*
 *   internal line source tracker 
 */
CHtmlDbg_src::CHtmlDbg_src(const textchar_t *fname, const textchar_t *path,
                           int source_id)
{
    /* remember the information */
    fname_.set(fname);
    source_id_ = source_id;
    
    /* store the path only if it's not empty */
    if (path != 0 && path[0] != '\0')
    {
        /* a non-empty path is provided - store it */
        path_.set(path);
    }
    else
    {
        /* 
         *   no path is provided - if the filename appears to be an
         *   absolute path, store it as the path 
         */
        if (os_is_file_absolute(fname))
            path_.set(fname);
    }
    
    /* not in a list yet */
    nxt_ = 0;
}

