#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/html/win32/mpegamp/mpegamp_w32.cpp,v 1.2 1999/05/17 02:52:26 MJRoberts Exp $";
#endif

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  mpegamp_w32.c - amp 0.7.6 audio driver for win95/98/nt
Function
  
Notes
  Derived from amp 0.7.6 for use in HTML TADS on Windows 95/98/NT.
Modified
  10/24/98 MJRoberts  - Creation
*/


#include <stdio.h>
#include <string.h>

#include <Windows.h>
#include <dsound.h>

#include "tadshtml.h"
#include "mpegamp_w32.h"

/* ------------------------------------------------------------------------ */
/*
 *   Statics 
 */
const int CMpegAmp::t_b8_l[2][3][22] =
{
    {
        {5,11,17,23,29,35,43,53,65,79,95,115,139,167,199,237,283,335,395,463,521,575},
        {5,11,17,23,29,35,43,53,65,79,95,113,135,161,193,231,277,331,393,463,539,575},
        {5,11,17,23,29,35,43,53,65,79,95,115,139,167,199,237,283,335,395,463,521,575}
    },
    {
        {3,7,11,15,19,23,29,35,43,51,61,73,89,109,133,161,195,237,287,341,417,575},
        {3,7,11,15,19,23,29,35,41,49,59,71,87,105,127,155,189,229,275,329,383,575},
        {3,7,11,15,19,23,29,35,43,53,65,81,101,125,155,193,239,295,363,447,549,575}
    }
};

const int CMpegAmp::t_b8_s[2][3][13] =
{
    {
        {3,7,11,17,23,31,41,55,73,99,131,173,191},
        {3,7,11,17,25,35,47,61,79,103,135,179,191},
        {3,7,11,17,25,35,47,61,79,103,133,173,191}
    },
    {
        {3,7,11,15,21,29,39,51,65,83,105,135,191},
        {3,7,11,15,21,27,37,49,63,79,99,125,191},
        {3,7,11,15,21,29,41,57,77,103,137,179,191}
    }
};

const short CMpegAmp::t_bitrate[2][3][15] =
{
    {
        {0,32,48,56,64,80,96,112,128,144,160,176,192,224,256},
        {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160},
        {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160}
    },
    {
        {0,32,64,96,128,160,192,224,256,288,320,352,384,416,448},
        {0,32,48,56,64,80,96,112,128,160,192,224,256,320,384},
        {0,32,40,48,56,64,80,96,112,128,160,192,224,256,320}
    }
};

const int CMpegAmp::t_sampling_frequency[2][3] =
{
    { 22050 , 24000 , 16000},
    { 44100 , 48000 , 32000}
};

/* ------------------------------------------------------------------------ */
/*
 *   Implementation 
 */


CMpegAmpW32::CMpegAmpW32(IDirectSound *ds)
{
    /* clear out 'this' */
    memset(this, 0, sizeof(*this));
    
    /* initialize the decoder */
    initialise_decoder();

    /* remember and reference our IDirectSound object */
    ds_ = ds;
    ds_->AddRef();

    /* no notification window yet */
    done_hwnd_ = 0;
    done_msg_ = 0;
    done_msg_lpar_ = 0;

    /* no playback thread yet */
    thread_hdl_ = 0;
    thread_id_ = 0;

    /* no URL yet */
    url_ = 0;
}

CMpegAmpW32::~CMpegAmpW32()
{
    /* wait for the playback thread to terminate */
    wait_until_done();
    
    /* make sure we've closed our buffer */
    audioBufferClose();
}

void CMpegAmp::printout(void)
{
    int j;

    if (nch == 2)
        j=32 * 18 * 2;
    else
        j=32 * 18;
    
    ((CMpegAmpW32 *)this)->
        audioBufferWrite((char*)sample_buffer, j * sizeof(short));
}

int CMpegAmpW32::audioBufferOpen(int freq, int stereo, int vol)
{
    DSBUFFERDESC bufdesc;

    /* no IDirectSound or IDirectSoundBuffer yet */
    dsbuf_ = 0;

    /* 
     *   configure the buffer based on the number of channels, the
     *   per-channel buffer size, and the MPEG version and layer 
     */
    nch = (stereo ? 2 : 1);
    data_size_ = nch * freq;

    /* we haven't sent anything to the sound card yet */
    flush_count_ = 0;

    /* set up our WAVEFORMATEX */
    wf_.wBitsPerSample  = 16;  // No 8-bit support yet
    wf_.wFormatTag = WAVE_FORMAT_PCM;
    wf_.nChannels = (WORD)nch;
    wf_.nSamplesPerSec = (DWORD)freq;
    wf_.nAvgBytesPerSec = (DWORD)nch * freq << 1;
    wf_.nBlockAlign = (WORD)(nch << 1);
    wf_.cbSize = 0;

    /* set up the direct sound buffer descriptor */
    memset(&bufdesc, 0, sizeof(bufdesc));
    bufdesc.dwSize = sizeof(bufdesc);
    bufdesc.dwFlags = DSBCAPS_CTRLDEFAULT
                      | DSBCAPS_GETCURRENTPOSITION2;
    bufdesc.dwBufferBytes = data_size_ * 3;
    bufdesc.dwReserved = 0;
    bufdesc.lpwfxFormat = &wf_;

    /* create our output buffer */
    if (ds_->CreateSoundBuffer(&bufdesc, &dsbuf_, 0) != DS_OK)
        return 2;

    /* set up to play from the beginning */
    dsbuf_->SetCurrentPosition(0);

    /* lock the initial buffer */
    dsbuf_->Lock(0, data_size_, &lock_ptr1_, &lock_len1_,
                 &lock_ptr2_, &lock_len2_, 0);
    stage_buf_cur_ = (char *)lock_ptr1_;
    stage_buf_end_ = stage_buf_cur_ + lock_len1_;

    /* move the write cursor past the first buffer */
    write_csr_ = data_size_;

    /* the direct sound buffer isn't playing yet */
    playing_ = FALSE;

    /* success */
    return 0;
}

/*
 *   begin playback if we haven't already done so 
 */
void CMpegAmpW32::directx_begin_play()
{
    if (!playing_)
    {
        dsbuf_->SetCurrentPosition(0);
        dsbuf_->Play(0, 0, DSBPLAY_LOOPING);
        playing_ = TRUE;
    }
}

/*
 *   Flush the staging buffer to directx 
 */
void CMpegAmpW32::flush_to_directx()
{
    HRESULT stat;

    /* unlock the current buffer */
    dsbuf_->Unlock(lock_ptr1_, lock_len1_, lock_ptr2_, lock_len2_);

    /* 
     *   if we're playing, wait for the playback pointer to move out of
     *   the buffer region we're going to attempt to write into 
     */
    if (playing_)
    {
        /* wait until playback moves out of our next write area */
        for (;;)
        {
            DWORD playpos;
            DWORD writepos;

            /* get the current playback position */
            dsbuf_->GetCurrentPosition(&playpos, &writepos);

            /* if it's in the region we want to write, wait a bit */
            if (playpos >= write_csr_ && playpos < write_csr_ + data_size_)
            {
                /* wait a bit, then check again */
                Sleep(100);
            }
            else
            {
                /* it's not in our playback region; we can proceed */
                break;
            }
        }
    }

    /*
     *   Attempt to lock one buffer-full at the current write position.
     *   This should always be the next position after our last write.  
     */
    for (;;)
    {
        /* try locking */
        stat = dsbuf_->Lock(write_csr_, data_size_,
                            &lock_ptr1_, &lock_len1_,
                            &lock_ptr2_, &lock_len2_, 0);

        /* check status */
        if (stat == DSERR_BUFFERLOST)
        {
            /* try restoring, and loop back for another try if successful */
            stat = dsbuf_->Restore();
            if (stat != DS_OK && stat != DSERR_BUFFERLOST)
                continue;

            /* failed to restore - give up and drop this buffer */
            return;
        }

        /* if we didn't get the buffer, give up and drop this buffer */
        if (stat != DS_OK)
            return;

        /* got it */
        break;
    }

    /* 
     *   increment our write cursor, wrapping if we reach the end of the
     *   direct sound buffer 
     */
    write_csr_ += data_size_;
    if (write_csr_ >= 3 * data_size_)
        write_csr_ = 0;

    /*
     *   Start playback after sending the first two buffers to directx.
     *   Wait until we've finished two buffers to help ensure that we keep
     *   a backlog in the pipe, so that the play cursor doesn't move into
     *   the part of the buffer we're working on filling.  
     */
    if (++flush_count_ == 2)
        directx_begin_play();

    /* 
     *   we've flushed out the staging buffer, so we can start filling it
     *   up again from the beginning 
     */
    stage_buf_cur_ = (char *)lock_ptr1_;
    stage_buf_end_ = stage_buf_cur_ + lock_len1_;
}

/*
 *   wait until we've played back the last buffer written 
 */
void CMpegAmpW32::wait_for_last_buf()
{
    DWORD playpos;
    DWORD writepos;
    DWORD last_write_csr;

    /* 
     *   if the underlying DirectSound buffer isn't playing, there's
     *   nothing to wait for 
     */
    if (!playing_)
        return;

    /* 
     *   note the current write cursor -- this is the start of the next
     *   buffer block after the last actual valid buffer block 
     */
    last_write_csr = write_csr_;

    /* 
     *   write out a buffer of silence, to ensure that we have a little
     *   time to catch it when the playback cursor moves out of our last
     *   buffer -- since we probably won't catch it at the exact instant
     *   it leaves, this will at least make it quiet until we manage to
     *   stop playback 
     */
    memset(lock_ptr1_, 0, stage_buf_end_ - (char *)lock_ptr1_);

    /* flush it */
    flush_to_directx();

    /* write out another buffer of silence */
    memset(lock_ptr1_, 0, stage_buf_end_ - (char *)lock_ptr1_);
    flush_to_directx();

    /* if anything is locked, unlock it */
    if (lock_ptr1_ != 0)
    {
        dsbuf_->Unlock(lock_ptr1_, lock_len1_, lock_ptr2_, lock_len2_);
        lock_ptr1_ = lock_ptr2_ = 0;
    }

    /* 
     *   keep waiting until the playback pointer moves into the next
     *   buffer after our last write buffer 
     */
    for (;;)
    {
        /* get the current playback position */
        dsbuf_->GetCurrentPosition(&playpos, &writepos);

        /* if it's not in the next region yet, wait a bit */
        if (playpos < last_write_csr
            || playpos >= last_write_csr + data_size_)
        {
            /* wait a bit, then check again */
            Sleep(100);
        }
        else
        {
            /* it's not in our playback region; we can proceed */
            break;
        }
    }

    /* stop playback */
    dsbuf_->Stop();
    playing_ = FALSE;
}

void CMpegAmpW32::audioBufferClose()
{
    /* if we never initialized, there's nothing to do */
    if (dsbuf_ == 0)
        return;

    /* check to see if we cancelled playback prematurely */
    if (stop_flag)
    {
        /* make sure playback is terminated */
        if (playing_)
        {
            dsbuf_->Stop();
            playing_ = FALSE;
        }
    }
    else
    {
        /* if we have a partially-filled buffer, write it out */
        if (stage_buf_cur_ != (char *)lock_ptr1_)
        {
            /* fill in the balance of the buffer with zeros */
            memset(stage_buf_cur_, 0, stage_buf_end_ - stage_buf_cur_);
            stage_buf_cur_ = stage_buf_end_;
            
            /* write it out */
            flush_to_directx();
        }

        /* wait for playback to finish */
        wait_for_last_buf();
    }

    /* release the DirectSound objects */
    if (dsbuf_ != 0)
        dsbuf_->Release();
    if (ds_ != 0)
        ds_->Release();

    /* remember that we've forgotten these */
    ds_ = 0;
    dsbuf_ = 0;
}

void CMpegAmpW32::audioBufferWrite(char *buf, int bytes)
{
    /* keep going until we run out of bytes to copy */
    while (bytes != 0)
    {
        DWORD copylen;
        DWORD max_copylen;
        
        /* copy as much as we can into the staging buffer */
        max_copylen = (DWORD)(stage_buf_end_ - stage_buf_cur_);
        copylen = (DWORD)bytes;
        if (copylen > max_copylen)
            copylen = max_copylen;

        /* copy these bytes */
        CopyMemory(stage_buf_cur_, buf, copylen);

        /* advance the pointers */
        stage_buf_cur_ += copylen;
        buf += copylen;

        /* deduct the amount copied from the remaining balance */
        bytes -= (int)copylen;

        /* we've filled up our staging buffer - flush it to directx */
        if (stage_buf_cur_ == stage_buf_end_)
            flush_to_directx();
    }
}


/* ------------------------------------------------------------------------ */
/*
 *   The remainder of this file is adapted from the original audio.c, the
 *   main entrypoint for the unix version 
 */

/* this file is a part of amp software, (C) tomislav uzelac 1996,1997
*/

/* audio.c      main amp source file 
 *
 * Created by: tomislav uzelac  Apr 1996 
 * Karl Anders Oygard added the IRIX code, 10 Mar 1997.
 * Ilkka Karvinen fixed /dev/dsp initialization, 11 Mar 1997.
 * Lutz Vieweg added the HP/UX code, 14 Mar 1997.
 * Dan Nelson added FreeBSD modifications, 23 Mar 1997.
 * Andrew Richards complete reorganisation, new features, 25 Mar 1997
 * Edouard Lafargue added sajber jukebox support, 12 May 1997
 */ 

int CMpegAmpW32::decodeMPEG(void)
{
    struct AUDIO_HEADER header;
    int cnt,g,snd_eof;

    initialise_globals();

    if ((g=gethdr(&header))!=0)
    {
        report_header_error(g);
        return -1;
    }
    
    if (header.protection_bit==0)
        getcrc();

    if (setup_audio(&header)!=0)
    {
        warn("Cannot set up audio. Exiting\n");
        return -1;
    }
        
    if (header.layer==1)
    {
        if (layer3_frame(&header,cnt)) {
            warn(" error. blip.\n");
            return -1;
        }
    }
    else if (header.layer==2)
    {
        if (layer2_frame(&header,cnt))
        {
            warn(" error. blip.\n");
            return -1;
        }
    }

    /*
     *   decoder loop ********************************** 
     */
    snd_eof=0;
    cnt=0;
    while (!snd_eof && !stop_flag)
    {
        while (!snd_eof)
        {
            if ((g=gethdr(&header))!=0)
            {
                report_header_error(g);
                snd_eof=1;
                break;
            }
            
            if (header.protection_bit==0)
                getcrc();
            
            if (header.layer==1)
            {
                if (layer3_frame(&header,cnt))
                {
                    warn(" error. blip.\n");
                    return -1;
                }
            }
            else if (header.layer==2)
            {
                if (layer2_frame(&header,cnt))
                {
                    warn(" error. blip.\n");
                    return -1;
                }
            }
            cnt++;
        }
    }
    return 0;
}

/* 
 *   call this once at the beginning 
 */
void CMpegAmp::initialise_decoder(void)
{
    premultiply();
    imdct_init();
    calculate_t43();
}

/* 
 *   call this before each file is played 
 */
void CMpegAmp::initialise_globals(void)
{
    append=data=nch=0; 
    f_bdirty=TRUE;
    bclean_bytes=0;
    
    memset(s,0,sizeof s);
    memset(res,0,sizeof res);
}

void CMpegAmp::report_header_error(int err)
{
    switch (err)
    {
    case GETHDR_ERR: warn("error reading mpeg bitstream.\n");
                     break;
    case GETHDR_NS : warn("MPEG 2.5 format is not supported -- this format "
                          "is not defined by the ISO/MPEG standard.");
                     break;
    case GETHDR_FL1: warn("ISO/MPEG layer 1 is not supported.\n");
                     break;
    case GETHDR_FF : warn("free format bitstreams are not supported.\n");
                     break;  
    case GETHDR_SYN: warn("error: out of sync.\n");
                     break;
    case GETHDR_EOF: 
    default:                ;   /* some stupid compilers need the semicolon */
    }       
}

/*
 *   set up the audio buffer 
 */
int CMpegAmpW32::setup_audio(struct AUDIO_HEADER *header)
{
    audioBufferOpen(t_sampling_frequency[header->ID]
                    [header->sampling_frequency],
                    (header->mode != 3 && !A_DOWNMIX),
                    A_SET_VOLUME);
    return 0;
}

/* ------------------------------------------------------------------------ */
/*
 *   Halt playback 
 */
void CMpegAmp::stop_playback()
{
    ((CMpegAmpW32 *)this)->directx_stop_play();
}

void CMpegAmpW32::directx_stop_play()
{
    /* if we're playing, stop directx */
    if (playing_)
    {
        dsbuf_->Stop();
        playing_ = FALSE;
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Show a warning 
 */
void CMpegAmp::warn(const char *msg, ...)
{
    va_list ap;

    if (((CMpegAmpW32 *)this)->url_ != 0)
        oshtml_dbg_printf("%s: ", ((CMpegAmpW32 *)this)->url_);
    
    va_start(ap, msg);
    oshtml_dbg_vprintf(msg, ap);
    va_end(ap);
}

/*
 *   dump the buffer 
 */
void CMpegAmp::dump(int *)
{
}

/* ------------------------------------------------------------------------ */
/*
 *   HTML TADS Interface 
 */

/*
 *   Play an MPEG file asynchronously
 */
int CMpegAmpW32::play(const textchar_t *url, HANDLE hfile, DWORD file_size,
                      HWND hwnd, UINT done_msg, LPARAM done_msg_lpar)
{
    /* remember the URL */
    url_ = url;
    
    /* remember the file handle and size */
    in_file = hfile;
    file_bytes_avail = file_size;

    /* remember the notification information */
    done_hwnd_ = hwnd;
    done_msg_ = done_msg;
    done_msg_lpar_ = done_msg_lpar;

    /* not stopping yet */
    stop_flag = FALSE;

    /* create the thread that will do the actual work */
    thread_hdl_ = CreateThread(0, 0,
                               (LPTHREAD_START_ROUTINE)play_thread_main,
                               (void *)this, 0, &thread_id_);
    if (thread_hdl_ == 0)
        return 1;

    /* success */
    return 0;
}

/*
 *   Wait until asynchronous playback is finished 
 */
void CMpegAmpW32::wait_until_done()
{
    /* if we have a background thread, wait for it */
    if (thread_hdl_ != 0)
    {
        /* wait for the playback thread to exit */
        WaitForSingleObject(thread_hdl_, INFINITE);
        
        /* we're done with the thread handle now */
        CloseHandle(thread_hdl_);
        thread_hdl_ = 0;
    }
}

/*
 *   Determine if playback is still underway.  Returns true if playback is
 *   active, false if playback has finished.  
 */
int CMpegAmpW32::is_playing()
{
    /* if we have no background thread, playback is done */
    if (thread_hdl_ == 0)
        return FALSE;

    /* wait for the thread, timing out immediately */
    switch(WaitForSingleObject(thread_hdl_, 0))
    {
    case WAIT_TIMEOUT:
        /* we're still playing */
        return TRUE;

    case WAIT_OBJECT_0:
        /* 
         *   it's finished - close the thread handle and indicate that
         *   playback has ended 
         */
        CloseHandle(thread_hdl_);
        thread_hdl_ = 0;
        return FALSE;

    default:
        /* anything else is an error; indicate that playback has finished */
        return FALSE;
    }
}


/*
 *   Playback thread main entrypoint
 */
void CMpegAmpW32::do_play_thread()
{
    /* decode the file */
    decodeMPEG();

    /* if playback was cancelled, stop directx */
    if (stop_flag && playing_)
    {
        dsbuf_->Stop();
        playing_ = FALSE;
    }

    /* close the audio device */
    audioBufferClose();

    /* 
     *   If we have a notification window, send it the message to let it
     *   know we're done.  Note that we never play back a sound more than
     *   once, so we'll always provide 1 for our repeat count.  
     */
    if (done_hwnd_ != 0)
        PostMessage(done_hwnd_, done_msg_, 1, done_msg_lpar_);
}

/*
 *   Stop playback 
 */
void CMpegAmpW32::stop(int sync)
{
    /* set the flag to indicate that we shouldn't do any more decoding */
    stop_flag = TRUE;

    /* if this is synchronous, wait until the playback thread exits */
    if (sync && thread_hdl_ != 0)
    {
        /* 
         *   wait for the playback thread to exit; give up if it doesn't
         *   exit after a few seconds 
         */
        WaitForSingleObject(thread_hdl_, 5000);

        /* we're done with the thread handle now */
        CloseHandle(thread_hdl_);
        thread_hdl_ = 0;
    }
}


#ifdef TEST_MPEGAMP
#include <conio.h>
/* ------------------------------------------------------------------------ */
/*
 *   Testing 
 */
void main(int argc,char **argv)
{
    CMpegAmpW32 *player;
    int argPos;
    IDirectSound *ds;
    HWND hwnd;

    /* create the IDirectSound interface */
    if (DirectSoundCreate(0, &ds, 0) != DS_OK)
    {
        printf("unable to create IDirectSound object\n");
        exit(2);
    }

    /* we need a window of some kind, so directx knows when we have focus */
    hwnd = CreateWindow("STATIC", "Hello!",
                        WS_POPUP | WS_VISIBLE,
                        100, 100, 200, 200, 0, 0, 0, 0);

    /* set the cooperative level */
    ds->SetCooperativeLevel(hwnd, DSSCL_NORMAL);

    /* create and initialize a player */
    player = new CMpegAmpW32(ds);
    player->initialise_decoder();

    /* check arguments */
    if (argc > 1)
    {
        /* play each file on the command line */
        for(argPos = 1 ; argPos < argc ; argPos++)
            player->play(argv[argPos], TRUE);
    }
    else
    {
        printf("usage: amp mp-file ...\n");
    }

    /* release our IDirectSound object */
    ds->Release();

    /* delete the player */
    delete player;

    /* delete our window */
    DestroyWindow(hwnd);

    exit(0);
}

/* 
 *   play back synchronously 
 */
void CMpegAmpW32::play(const char *inFileStr, int stop_on_key)
{
    HANDLE hfile;
    DWORD file_size;
    int err;

    /* open the file */
    hfile = CreateFile(inFileStr, GENERIC_READ, FILE_SHARE_READ,
                         0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    if (in_file == INVALID_HANDLE_VALUE)
    {
        warn("Could not open file: %s\n", inFileStr);
        return;
    }

    /* get the file's size */
    file_size = GetFileSize(hfile, 0);

    /* start the playback thread */
    err = play(hfile, file_size, 0, 0, 0);
    if (err != 0)
    {
        printf("error %d from play()\n", err);
        goto done;
    }

    /* wait according to the caller's specifications */
    if (stop_on_key)
    {
        /* wait until playback is finished or the user presses a key */
        for (;;)
        {
            /* if playback is finished, stop */
            if (!is_playing())
                break;
            
            /* if the user has pressed a key, stop */
            if (_kbhit())
            {
                /* stop playback */
                stop(TRUE);
                
                /* no need to wait any longer */
                break;
            }
        }
    }
    else
    {
        /* wait until the thread terminates */
        wait_until_done();
    }

done:
    /* close the file */
    CloseHandle(in_file);
}

#endif

